from typing import Dict, Optional

from pydantic import BaseModel

from dictum_core.schema.id import ID
from dictum_core.schema.model.calculations import (
    DetachedDimension,
    DimensionsUnion,
    Metric,
)
from dictum_core.schema.model.table import Table
from dictum_core.schema.model.transform import Transform

root_keys = {"tables", "metrics", "unions"}


class Model(BaseModel):
    name: str
    description: Optional[str] = None
    locale: str = "en_US"
    currency: str = "USD"

    dimensions: Dict[ID, DetachedDimension] = {}
    metrics: Dict[ID, Metric] = {}
    unions: Dict[ID, DimensionsUnion] = {}

    tables: Dict[ID, Table] = {}
    transforms: Dict[
        ID, Transform
    ] = {}  # ignored for now, TODO: load as LiteralTransform

    theme: Optional[dict] = None



from dictum_core import Project
from dictum_core.project.yaml_mapped_dict import YAMLMappedDict


def generate():
    project = Project.example("chinook")
    project.model_data = YAMLMappedDict(
        {"name": "Chinook", "tables": {}, "metrics": {}, "unions": {}}
    )
    project.project_config.root = None
    project.update_model({})
    return project



import pytest

from dictum_core.model.types import Type, grains, resolve_type


def test_resolve_basic_types():
    assert resolve_type("bool") == Type(name="bool")
    assert resolve_type("str") == Type(name="str")
    assert resolve_type("int") == Type(name="int")
    assert resolve_type("float") == Type(name="float")


def test_resolve_basic_datetime_types():
    assert resolve_type("datetime") == Type(name="datetime", grain="second")
    assert resolve_type("date") == Type(name="datetime", grain="day")


def test_resolve_grains():
    for grain in grains:
        assert resolve_type(f"datetime:{grain}") == Type(name="datetime", grain=grain)
        assert resolve_type(f"date:{grain}") == Type(name="datetime", grain=grain)


def test_resolve_invalid_grain():
    with pytest.raises(ValueError, match=r"Unknown datetime grain: invalid"):
        resolve_type("datetime:invalid")


def test_resolve_grain_invalid_type():
    with pytest.raises(
        ValueError,
        match=(
            r"Time grains can be specified only for date and datetime types,"
            " got grain for int"
        ),
    ):
        resolve_type("int:day")


def test_resolve_unknown_type():
    with pytest.raises(ValueError, match=r"Unknown type: unknown"):
        resolve_type("unknown")


def test_type_grains():
    T = resolve_type("date:month")
    assert T.grains == ["year", "quarter", "month"]

    assert resolve_type("int").grains == []



from dictum_core.format import Format
from dictum_core.model.scalar import transforms_by_input_type
from dictum_core.model.types import Type

datetime_transforms = transforms_by_input_type["datetime"]


class GenericTimeDimension:
    type: Type
    grain: str

    _sort_counter = 0

    def __init__(self, locale: str):
        self.locale = locale

    def __init_subclass__(cls):
        cls.id = cls.__name__
        cls.sort_order = GenericTimeDimension._sort_counter
        GenericTimeDimension._sort_counter += 1

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return str(self)

    @property
    def name(self) -> str:
        return str(self)

    @property
    def format(self) -> Format:
        return Format(locale=self.locale, type=self.type)

    @property
    def type(self) -> Type:
        return Type(name="datetime", grain=self.grain)

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    @property
    def transforms(self) -> dict:
        return datetime_transforms


class Time(GenericTimeDimension):
    grain = "second"


class Year(GenericTimeDimension):
    grain = "year"


class Quarter(GenericTimeDimension):
    grain = "quarter"


class Month(GenericTimeDimension):
    grain = "month"


class Week(GenericTimeDimension):
    grain = "week"


class Day(GenericTimeDimension):
    grain = "day"


class Date(Day):
    pass


class Hour(GenericTimeDimension):
    grain = "hour"


class Minute(GenericTimeDimension):
    grain = "minute"


class Second(GenericTimeDimension):
    grain = "second"


dimensions = {
    "Time": Time,
    "Year": Year,
    "Quarter": Quarter,
    "Month": Month,
    "Week": Week,
    "Day": Day,
    "Date": Date,
    "Hour": Hour,
    "Minute": Minute,
    "Second": Second,
}



import json
from pathlib import Path

from jinja2.environment import Environment
from jinja2.loaders import FileSystemLoader

path = Path(__file__).parent
environment = Environment(loader=FileSystemLoader(searchpath=path))

lineage_spec = json.loads((path / "lineage.json").read_text())



from pathlib import Path
from textwrap import dedent

import pytest
import yaml
from lark import Tree

from dictum_core.ql import compile_query
from dictum_core.ql.parser import (
    parse_dimension,
    parse_dimension_request,
    parse_metric_request,
    parse_ql,
)
from dictum_core.ql.transformer import (
    compile_dimension,
    compile_dimension_request,
    compile_metric_request,
)
from dictum_core.schema.query import (
    Query,
    QueryDimension,
    QueryDimensionRequest,
    QueryMetricRequest,
)


def _assert_tree(tree: Tree, expected: str):
    assert tree.pretty().strip() == dedent(expected).strip()


with (Path(__file__).parent / "test_ql_parse.yml").open("r") as fp:
    parse_cases = yaml.safe_load(fp)


@pytest.mark.parametrize("case", parse_cases)
def test_ql_parse(case: dict):
    _kind_mapping = {
        "dimension": parse_dimension,
        "dimension_request": parse_dimension_request,
        "metric_request": parse_metric_request,
        "query": parse_ql,
    }
    val = case["val"]
    expected = case["expected"]
    kind = case["kind"]
    func = _kind_mapping[kind]
    assert func(val).pretty().strip() == dedent(expected.strip())


def test_compile_query():
    assert compile_query("select x") == Query.model_validate(
        {"metrics": [{"metric": {"id": "x"}}]}
    )


def test_compile_groupby():
    q = Query.model_validate(
        {
            "metrics": [{"metric": {"id": "x"}}],
            "dimensions": [{"dimension": {"id": "y"}}],
        }
    )
    assert compile_query("select x group by y") == q
    assert compile_query("select x by y") == q


def test_compile_groupby_transforms():
    assert compile_query("select x, y by z.p(10)") == Query.model_validate(
        {
            "metrics": [{"metric": {"id": "x"}}, {"metric": {"id": "y"}}],
            "dimensions": [
                {"dimension": {"id": "z", "transforms": [{"id": "p", "args": [10]}]}}
            ],
        }
    )


def test_compile_where():
    assert compile_query("select x where y.z(10)") == Query.model_validate(
        {
            "metrics": [{"metric": {"id": "x"}}],
            "filters": [{"id": "y", "transforms": [{"id": "z", "args": [10]}]}],
        }
    )


def test_compile_multiple_groupbys():
    compile_query(
        """
    select x, y
    where z.z(1, 2, 3),
        a.b('a')
    by d.d('d'),
       c,
       f.h(11)
    """
    ) == Query.model_validate(
        {
            "metrics": [{"metric": {"id": "x"}}, {"metric": {"id": "y"}}],
            "dimensions": [
                {
                    "dimension": {
                        "id": "d",
                        "transforms": [{"id": "d", "args": ["d"]}],
                    },
                },
                {"dimension": {"id": "c"}},
                {"dimension": {"id": "f", "tranforms": [{"id": "h", "args": [11]}]}},
            ],
            "filters": [
                {
                    "id": "z",
                    "transforms": [{"id": "z", "args": [1, 2, 3]}],
                },
                {"id": "a", "transforms": [{"id": "b", "args": ["a"]}]},
            ],
        }
    )


def test_compile_dimension_alias():
    assert compile_query("select metric by dim as dim1") == Query.model_validate(
        {
            "metrics": [{"metric": {"id": "metric"}}],
            "dimensions": [{"dimension": {"id": "dim"}, "alias": "dim1"}],
        }
    )


def test_compile_dimension_transform_alias():
    assert compile_query("select x by y.a.b(10) as z") == Query.model_validate(
        {
            "metrics": [{"metric": {"id": "x"}}],
            "dimensions": [
                {
                    "dimension": {
                        "id": "y",
                        "transforms": [{"id": "a"}, {"id": "b", "args": [10]}],
                    },
                    "alias": "z",
                }
            ],
        }
    )


def test_compile_dimension_request():
    assert compile_dimension_request("x") == QueryDimensionRequest.model_validate(
        {"dimension": {"id": "x"}}
    )
    assert compile_dimension_request("x.y") == QueryDimensionRequest.model_validate(
        {"dimension": {"id": "x", "transforms": [{"id": "y"}]}}
    )
    assert compile_dimension_request(
        "x.y('z', 1) as a"
    ) == QueryDimensionRequest.model_validate(
        {
            "dimension": {"id": "x", "transforms": [{"id": "y", "args": ["z", 1]}]},
            "alias": "a",
        }
    )


def test_compile_metric_request():
    assert compile_metric_request(
        "x.y(1, 'f') of (a) within (b, c.n(1)) as al"
    ) == QueryMetricRequest.model_validate(
        {
            "metric": {
                "id": "x",
                "transforms": [
                    {
                        "id": "y",
                        "args": [1, "f"],
                        "of": [{"id": "a"}],
                        "within": [
                            {"id": "b"},
                            {"id": "c", "transforms": [{"id": "n", "args": [1]}]},
                        ],
                    }
                ],
            },
            "alias": "al",
        }
    )


def test_compile_filter():
    assert compile_dimension("x.y('z')") == QueryDimension.model_validate(
        {"id": "x", "transforms": [{"id": "y", "args": ["z"]}]}
    )


def test_compile_filter_null():
    assert compile_dimension("x is null") == QueryDimension.model_validate(
        {"id": "x", "transforms": [{"id": "isnull", "args": []}]}
    )



from dataclasses import dataclass
from typing import List, Literal, Optional

grain_grains = {
    "year": ["year"],
    "quarter": ["year", "quarter"],
    "month": ["year", "quarter", "month"],
    "week": ["year", "week"],
    "day": ["year", "quarter", "month", "week", "day"],
    "hour": ["year", "quarter", "month", "week", "day", "hour"],
    "minute": ["year", "quarter", "month", "week", "day", "hour", "minute"],
    "second": ["year", "quarter", "month", "week", "day", "hour", "minute", "second"],
}


grains = set(grain_grains)


@dataclass
class Type:
    name: Literal["bool", "int", "float", "datetime", "str"]
    grain: Optional[
        Literal["year", "quarter", "month", "week", "day", "hour", "minute", "second"]
    ] = None

    @property
    def grains(self) -> List[str]:
        if self.name != "datetime":
            return []
        return grain_grains[self.grain]


def resolve_type(name: str) -> Type:
    if name in {"bool", "int", "float", "str"}:
        return Type(name=name)
    if name == "date":
        return Type(name="datetime", grain="day")
    if name == "datetime":
        return Type(name="datetime", grain="second")
    if ":" in name:
        _type, grain = name.split(":", maxsplit=1)
        if grain not in grains:
            raise ValueError(f"Unknown datetime grain: {grain}")
        if _type not in {"datetime", "date"}:
            raise ValueError(
                "Time grains can be specified only for date and datetime types,"
                f" got grain for {_type}"
            )
        return Type(name="datetime", grain=grain)
    raise ValueError(f"Unknown type: {name}")



import dataclasses
from typing import Any, Dict, List, Optional, Union

from dictum_core import schema
from dictum_core.format import Format
from dictum_core.model.calculations import (
    Calculation,
    Dimension,
    DimensionsUnion,
    Expression,
    Measure,
    Metric,
    TableCalculation,
)
from dictum_core.model.scalar import transforms as scalar_transforms
from dictum_core.model.table import Table, TableFilter
from dictum_core.model.time import GenericTimeDimension
from dictum_core.model.time import dimensions as time_dimensions
from dictum_core.model.types import resolve_type

displayed_fields = {"name", "description", "missing"}

table_calc_fields = displayed_fields | {"str_expr"}


class Model:
    def __init__(self, name: Optional[str] = None) -> None:
        self.name = name
        self.description = None
        self.locale = "en_US"
        self.scalar_transforms = scalar_transforms
        self.theme = None

        self.tables: Dict[str, Table] = {}
        self.measures: Dict[str, Measure] = {}
        self.dimensions: Dict[str, Dimension] = {}
        self.metrics: Dict[str, Metric] = {}

    @classmethod
    def from_config(cls, model: schema.Model):
        from dictum_core.model.checks import check_model
        from dictum_core.schema.model.checks import check_model_config

        check_model_config(model)

        obj = cls(name=model.name)
        obj.description = model.description
        obj.locale = model.locale
        obj.theme = model.theme

        # add unions
        for id_, union in model.unions.items():
            union_type = resolve_type(union.type)
            obj.dimensions[id_] = DimensionsUnion(
                **union.model_dump(include=displayed_fields),
                id=id_,
                format=Format(
                    type=union_type,
                    config=union.format,
                    locale=obj.locale,
                    default_currency=model.currency,
                ),
                type=union_type,
            )

        # add all tables, their relationships and calculations
        for id_, config_table in model.tables.items():
            table = obj.add_table(
                **config_table.model_dump(
                    include={"source", "description", "primary_key", "filters"}
                ),
                id=id_,
            )

            for alias, related in config_table.related.items():
                table.add_related(
                    str_table=related.str_table,
                    related_key=related.str_related_key,
                    foreign_key=related.foreign_key,
                    alias=alias,
                    tables=obj.tables,
                )

            # add table dimensions
            for id_, dimension in config_table.dimensions.items():
                obj.add_dimension(
                    id=id_,
                    table=table.id,
                    format_config=dimension.format,
                    type=dimension.type,
                    currency=model.currency,
                    **dimension.model_dump(include=table_calc_fields | {"union"}),
                )

            # add table measures
            for id_, measure in config_table.measures.items():
                obj.add_measure(
                    table=table.id,
                    id=id_,
                    format_config=measure.format,
                    type=measure.type,
                    currency=model.currency,
                    **measure.model_dump(
                        include=table_calc_fields | {"str_filter", "str_time"}
                    ),
                )

        # add detached dimensions
        # for dimension in model.dimensions.values():
        #     table = obj.tables[dimension.table]
        #     obj.add_dimension(dimension, table)

        # add metrics
        for id_, metric in model.metrics.items():
            obj.add_metric(
                id=id_,
                type=metric.type,
                format_config=metric.format,
                currency=model.currency,
                **metric.model_dump(
                    include=table_calc_fields | {"str_filter", "str_time", "table"}
                ),
            )

        # add measure backlinks
        for table in obj.tables.values():
            for measure in table.measures.values():
                for target in table.allowed_join_paths:
                    target.measure_backlinks[measure.id] = table

        # add default time dimensions
        for id_, time_dimension in time_dimensions.items():
            obj.dimensions[id_] = time_dimension(locale=obj.locale)

        check_model(obj)
        return obj

    def add_table(
        self,
        id: str,
        source: Union[str, Dict[str, str]],
        description: Optional[str] = None,
        primary_key: Optional[str] = None,
        filters: Optional[List[str]] = None,
    ) -> Table:
        result = Table(
            id=id, source=source, description=description, primary_key=primary_key
        )
        self.tables[result.id] = result
        if filters:
            result.filters = [TableFilter(str_expr=f, table=result) for f in filters]
        return result

    def add_measure(
        self,
        table: str,
        id: str,
        name: str,
        str_expr: str,
        currency: str,
        description: Optional[str] = None,
        type: Optional[str] = None,
        missing: Optional[Any] = None,
        format_config: Optional[schema.FormatConfig] = None,
        str_filter: Optional[str] = None,
        str_time: Optional[str] = None,
        metric: Optional[bool] = True,
    ) -> Measure:
        if type is None:
            type = "float"
        type = resolve_type(type)
        table = self.tables.get(table)
        result = Measure(
            model=self,
            table=table,
            id=id,
            name=name,
            description=description,
            type=type,
            missing=missing,
            str_expr=str_expr,
            str_filter=str_filter,
            str_time=str_time,
            format=Format(
                type=type,
                config=format_config,
                locale=self.locale,
                default_currency=currency,
            ),
        )
        if metric:
            metric_ = Metric.from_measure(result, self)
            self.metrics[metric_.id] = metric_
        table.measures[result.id] = result
        self.measures[result.id] = result
        return result

    def add_dimension(
        self,
        table: str,
        id: str,
        name: str,
        str_expr: str,
        type: str,
        currency: str,
        description: Optional[str] = None,
        missing: Optional[Any] = None,
        format_config: Optional[schema.FormatConfig] = None,
        union: Optional[str] = None,
    ) -> Dimension:
        table = self.tables.get(table)
        type = resolve_type(type)
        result = Dimension(
            table=table,
            id=id,
            name=name,
            str_expr=str_expr,
            description=description,
            missing=missing,
            type=type,
            format=Format(
                type=type,
                config=format_config,
                locale=self.locale,
                default_currency=currency,
            ),
        )
        table.dimensions[result.id] = result
        if union is not None:
            if union in table.dimensions:
                raise KeyError(
                    f"Duplicate union dimension {union} " f"on table {table.id}"
                )
            union: DimensionsUnion = self.dimensions.get(union)
            table.dimensions[union.id] = dataclasses.replace(
                result,
                id=union.id,
                name=union.name,
                description=union.description,
                type=union.type,
                format=union.format,
                missing=union.missing,
                is_union=True,
            )
        self.dimensions[result.id] = result
        return result

    def add_metric(
        self,
        id: str,
        name: str,
        str_expr: str,
        currency: str,
        description: Optional[str] = None,
        missing: Optional[Any] = None,
        type: Optional[str] = None,
        format_config: Optional[schema.FormatConfig] = None,
        str_filter: Optional[str] = None,
        str_time: Optional[str] = None,
        table: Optional[str] = None,
    ):
        if table is not None:
            # table is specified, treat as that table's measure
            return self.add_measure(
                table=table,
                id=id,
                name=name,
                str_expr=str_expr,
                description=description,
                missing=missing,
                type=type,
                format_config=format_config,
                str_time=str_time,
                str_filter=str_filter,
                currency=currency,
                metric=True,
            )

        # no, it's a real metric
        if type is None:
            type = "float"
        type = resolve_type(type)
        result = Metric(
            model=self,
            id=id,
            name=name,
            str_expr=str_expr,
            description=description,
            missing=missing,
            type=type,
            format=Format(
                locale=self.locale,
                type=type,
                config=format_config,
                default_currency=currency,
            ),
        )
        self.metrics[result.id] = result
        return result

    def get_lineage(
        self, calculation: Calculation, parent: Optional[str] = None
    ) -> List[dict]:
        """Get lineage graph for a calculation. Returns a list of dicts like:

        {"id": "someid", "type": "Metric", "parent": "otherid"}
        """
        _id = f"{calculation.__class__.__name__}:{calculation.id}"
        yield {
            "id": _id,
            "name": calculation.id,
            "parent": parent,
            "type": calculation.__class__.__name__,
        }
        expr = calculation.parsed_expr
        table = None
        has_refs = False
        for k in ("metric", "measure", "dimension"):
            attr = f"{k}s"
            for ref in expr.find_data(k):
                has_refs = True
                yield from self.get_lineage(
                    getattr(self, attr)[ref.children[0]], parent=_id
                )
        if isinstance(calculation, TableCalculation):
            table = calculation.table.id
            prefix = [table] if table is not None else []
            for ref in expr.find_data("column"):
                has_refs = True
                _col = ".".join([*prefix, *ref.children])
                yield {
                    "id": _col,
                    "type": "Column",
                    "parent": _id,
                    "name": _col,
                }
            if not has_refs:
                yield {
                    "id": table,
                    "type": "Column",
                    "parent": _id,
                    "name": f"{table}.*",
                }

    @property
    def expressions(self) -> List[Expression]:
        """List all child objects that are instances of Expression.
        Used in checks.
        """
        yield from self.metrics.values()
        yield from (
            d
            for d in self.dimensions.values()
            if not isinstance(d, GenericTimeDimension)
            and not isinstance(d, DimensionsUnion)
        )
        for table in self.tables.values():
            yield from table.filters
        for measure in self.measures.values():
            yield measure
            if measure.filter is not None:
                yield measure.filter

    # def suggest_metrics(self, query: schema.Query) -> List[Measure]:
    #     """Suggest a list of possible metrics based on a query.
    #     Only metrics that can be used with all the dimensions from the query
    #     """
    #     result = []
    #     query_dims = set(r.dimension.id for r in query.dimensions)
    #     for metric in self.metrics.values():
    #         if metric.id in query.metrics:
    #             continue
    #         allowed_dims = set(d.id for d in metric.dimensions)
    #         if query_dims < allowed_dims:
    #             result.append(metric)
    #     return sorted(result, key=lambda x: (x.name))

    # def suggest_dimensions(self, query: schema.Query) -> List[Dimension]:
    #     """Suggest a list of possible dimensions based on a query. Only dimensions
    #     shared by all measures that are already in the query.
    #     """
    #     dims = set(self.dimensions) - set(r.dimension.id for r in query.dimensions)
    #     for request in query.metrics:
    #         metric = self.metrics.get(request.metric.id)
    #         for measure in metric.measures:
    #             dims = dims & set(measure.table.allowed_dimensions)
    #     return sorted([self.dimensions[d] for d in dims], key=lambda x: x.name)

    # def get_range_computation(self, dimension_id: str) -> Computation:
    #     """Get a computation that will compute a range of values for a given
    #       dimension.
    #     This is seriously out of line with what different parts of computation mean,
    #     so maybe we need to give them more abstract names.
    #     """
    #     dimension = self.dimensions.get(dimension_id)
    #     table = dimension.table
    #     min_, max_ = dimension.prepare_range_expr([table.id])
    #     return Computation(
    #         queries=[
    #             AggregateQuery(
    #                 join_tree=AggregateQuery(table=table, identity=table.id),
    #                 aggregate={"min": min_, "max": max_},
    #             )
    #         ]
    #     )

    # def get_values_computation(self, dimension_id: str) -> Computation:
    #     """Get a computation that will compute a list of unique possible values for
    #     this dimension.
    #     """
    #     dimension = self.dimensions.get(dimension_id)
    #     table = dimension.table
    #     return Computation(
    #         queries=[
    #             AggregateQuery(
    #                 join_tree=AggregateQuery(table=table, identity=table.id),
    #                 groupby={"values": dimension.prepare_expr([table.id])},
    #             )
    #         ]
    #     )

    def get_currencies_for_query(self, query: schema.Query):
        currencies = set()
        for request in query.metrics:
            metric = self.metrics.get(request.metric.id)
            if metric.format.currency is not None:
                currencies.add(metric.format.currency)
        for request in query.dimensions:
            dimension = self.dimensions.get(request.dimension.id)
            if dimension.format.currency is not None:
                currencies.add(dimension.format.currency)
        return currencies



import inspect
import shutil
from pathlib import Path

import yaml
from jinja2 import Template

from dictum_core.backends.base import Backend
from dictum_core.backends.secret import Secret

template_path = Path(__file__).parent / "project_template"


def _get_backend_parameters(backend: Backend) -> dict:
    result = {}
    for name, par in inspect.signature(backend.__init__).parameters.items():
        if name == "self":
            continue
        if name not in backend.parameters:
            raise KeyError(
                f"Missing parameter {name} in {backend.type} backend parameters"
            )
        if par.annotation is Secret:
            var_name = f"DICTUM_SECRET_{backend.type}_{name}".upper()
            result[name] = "{{ env.%s }}" % var_name
        else:
            result[name] = backend.parameters[name]
    return result


def copy_project_template(target_path: Path, template_vars: dict):
    for path in template_path.iterdir():
        if path.name in {"__init__.py", ".gitkeep"}:
            continue
        new_path = target_path / path.relative_to(template_path)
        if new_path.exists():
            raise FileExistsError(f"{new_path} already exists")
        if path.is_file():
            template = Template(path.read_text())
            rendered = template.render(**template_vars)
            new_path.write_text(rendered)
        elif path.is_dir():
            shutil.copytree(str(path), str(new_path))


def create_new_project(
    path: Path,
    backend: Backend,
    name: str,
    currency: str = "USD",
    locale: str = "en_US",
):
    if not path.parent.exists():
        raise FileNotFoundError(f"{path.parent} directory doesn't exist")
    if path.is_file():
        raise FileExistsError(f"{path} is file")
    if not path.exists():
        path.mkdir()
    if path.is_dir():
        template_vars = {
            "project_name": name,
            "profile": "default",
            "backend": backend.type,
            "backend_parameters": yaml.safe_dump(_get_backend_parameters(backend)),
            "currency": currency,
            "locale": locale,
        }
        copy_project_template(path, template_vars)
        return
    raise Exception("This shouldn't happen")



from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field

from dictum_core.schema.id import ID
from dictum_core.schema.model.calculations import Dimension, Measure


class RelatedTable(BaseModel):
    str_table: str = Field(alias="table")
    foreign_key: str
    str_related_key: Optional[str] = Field(None, alias="related_key")


class Table(BaseModel):
    description: Optional[str] = None
    source: Union[str, Dict]
    primary_key: Optional[str] = None
    related: Dict[ID, RelatedTable] = {}
    measures: Dict[ID, Measure] = {}
    dimensions: Dict[ID, Dimension] = {}
    filters: List[str] = []



from functools import wraps

from lark import Lark, Token, Transformer, Tree
from lark.exceptions import UnexpectedInput

from dictum_core import grammars
from dictum_core.exceptions import ShorthandSyntaxError

grammars = grammars.__file__

related_parser = Lark.open("magics.lark", rel_to=grammars, start="related")
metric_parser = Lark.open(
    "magics.lark", rel_to=grammars, start="metric", propagate_positions=True
)
dimension_parser = Lark.open(
    "magics.lark", rel_to=grammars, start="dimension", propagate_positions=True
)
table_parser = Lark.open(
    "magics.lark", rel_to=grammars, start="table_full", propagate_positions=True
)
format_parser = Lark.open("magics.lark", rel_to=grammars, start="format")


class Preprocessor(Transformer):
    def IDENTIFIER(self, token: Token):
        return token.value

    def identifier(self, children: list):
        return children[0]

    def key_value(self, children: list):
        return tuple(children)

    def key_values(self, children: list):
        return dict(children)

    def ql__IDENTIFIER(self, token: Token):
        return token.value.strip('"')

    def ql__QUOTED_IDENTIFIER(self, token: Token):
        return token.value.strip('"')

    def table_metric(self, children: list):
        return children[0]

    def table_dimension(self, children: list):
        return children[0]

    def table_related(self, children: list):
        return children[0]

    def filter(self, children: list):
        """Inline expr node directly into filter"""
        return Tree("filter", children[0].children, meta=children[0].meta)


preprocessor = Preprocessor()


def catch_syntax_errors(fn: callable):
    @wraps(fn)
    def wrapped(definition: str):
        try:
            return fn(definition)
        except UnexpectedInput as e:
            raise ShorthandSyntaxError(e, definition)

    return wrapped


@catch_syntax_errors
def parse_shorthand_table(definition: str) -> Tree:
    return preprocessor.transform(table_parser.parse(definition))


@catch_syntax_errors
def parse_shorthand_metric(definition: str) -> Tree:
    return preprocessor.transform(metric_parser.parse(definition))


@catch_syntax_errors
def parse_shorthand_dimension(definition: str) -> Tree:
    return preprocessor.transform(dimension_parser.parse(definition))


@catch_syntax_errors
def parse_shorthand_related(definition: str) -> Tree:
    return preprocessor.transform(related_parser.parse(definition))


@catch_syntax_errors
def parse_shorthand_format(definition: str) -> Tree:
    return preprocessor.transform(format_parser.parse(definition))



from pathlib import Path
from textwrap import dedent

import pytest
import yaml
from lark import Tree, UnexpectedCharacters

from dictum_core.exceptions import MixedExpressionError
from dictum_core.model.expr.introspection import get_expr_kind, get_expr_total_function
from dictum_core.model.expr.parser import parse_expr

with (Path(__file__).parent / "test_expr_parse.yml").open("r") as fp:
    parse_cases = yaml.safe_load(fp)


@pytest.mark.parametrize("case", parse_cases)
def test_parse(case: dict):
    val = case["val"]
    expected = case["expected"]
    missing = case.get("missing")
    assert parse_expr(val, missing=missing).pretty().strip() == dedent(expected.strip())


def test_measure():
    """Referencing other calculation"""
    assert parse_expr("$calc").children[0] == Tree("measure", ["calc"])
    with pytest.raises(UnexpectedCharacters):
        parse_expr("$table.calc")


def test_dimension():
    """Referencing other dimensions."""
    assert parse_expr(":dim").children[0] == Tree("dimension", ["dim"])
    with pytest.raises(UnexpectedCharacters):
        parse_expr(":table.dim")


def test_expr_kind():
    assert get_expr_kind(parse_expr("1")) == "scalar"
    assert get_expr_kind(parse_expr("'a'")) == "scalar"

    assert get_expr_kind(parse_expr("col")) == "column"
    assert get_expr_kind(parse_expr("col + 1")) == "column"
    assert get_expr_kind(parse_expr("col * 2")) == "column"

    assert get_expr_kind(parse_expr("sum(col)")) == "aggregate"
    assert get_expr_kind(parse_expr("count(col) - 1")) == "aggregate"
    assert get_expr_kind(parse_expr("min(col) * 100")) == "aggregate"
    assert get_expr_kind(parse_expr("max(col) / 100")) == "aggregate"
    assert get_expr_kind(parse_expr("floor(sum(col - 1) + 1) / 10")) == "aggregate"
    assert get_expr_kind(parse_expr("case when max(col) > 1 then 1 else 0 end"))

    assert get_expr_kind(parse_expr("sum(col) in (1, 2, 3)")) == "aggregate"
    assert get_expr_kind(parse_expr("col in (1, 2, 3)")) == "column"
    assert get_expr_kind(parse_expr("1 in (1, 2, 3)")) == "scalar"


def test_expr_kind_invalid():
    with pytest.raises(MixedExpressionError):
        get_expr_kind(parse_expr("sum(col) + col"))
    with pytest.raises(MixedExpressionError):
        get_expr_kind(parse_expr("case when col = 1 then sum(col) end"))
    with pytest.raises(MixedExpressionError):
        get_expr_kind(parse_expr("case when col = 1 then sum(col) end"))
    with pytest.raises(MixedExpressionError):
        get_expr_kind(parse_expr("col in (sum(col), 1, 2)"))


def test_expr_total_function():
    assert get_expr_total_function(parse_expr("sum(col)")) == "sum"
    assert get_expr_total_function(parse_expr("sum(col + 1)")) == "sum"
    assert get_expr_total_function(parse_expr("count(col)")) == "sum"
    assert get_expr_total_function(parse_expr("max(col)")) == "max"
    assert get_expr_total_function(parse_expr("min(col)")) == "min"
    assert get_expr_total_function(parse_expr("avg(col)")) is None
    assert get_expr_total_function(parse_expr("countd(col)")) is None
    assert get_expr_total_function(parse_expr("sum(col) + 1")) is None


def test_mixed_bug():
    get_expr_kind(parse_expr("count(case when type = 1 then 1 end) / count()"))



import pytest

from dictum_core import schema
from dictum_core.exceptions import (
    AggregateDimensionError,
    CircularReferenceError,
    ExpressionSyntaxError,
    FilteredMeasureReferenceError,
    InvalidMeasureError,
    InvalidReferenceError,
    MissingAggregationError,
    MissingDimensionError,
    MissingMeasureError,
    MissingPrimaryKeyError,
    MixedExpressionError,
    NonAggregateMeasureError,
    UnknownFunctionError,
)
from dictum_core.model.checks import check_model
from dictum_core.model.model import Model
from dictum_core.schema.model.checks import check_model_config


def build_model(data: dict) -> Model:
    config = schema.Model.model_validate(data)
    return Model.from_config(config)


@pytest.fixture(scope="function")
def model() -> Model:
    """Provides a valid model that can be broken for tests"""
    config = schema.Model.model_validate(
        {
            "name": "test",
            "tables": {
                "t1": {
                    "source": "t1",
                    "primary_key": "pk",
                    "measures": {
                        "me1": {"name": "me1", "expr": "sum(col)"},
                        "me3": {"name": "me3", "expr": "sum(col)"},
                    },
                    "dimensions": {
                        "d1": {"name": "d1", "expr": "col", "type": "int"},
                        "d2": {"name": "d2", "expr": "$m1", "type": "int"},
                    },
                    "filters": [":d1 > 0"],
                },
                "t2": {
                    "source": "t2",
                    "measures": {"me2": {"name": "me2", "expr": "count()"}},
                },
            },
            "metrics": {
                "m1": {
                    "name": "m1",
                    "expr": "sum(value)",
                    "table": "t1",
                    "filter": ":d1 = 1 and col = 2",
                },
                "m2": {"name": "m2", "expr": "$m1 / 100"},
                "m3": {"name": "m3", "expr": "$m1 / $m2"},
            },
        }
    )
    check_model_config(config)
    return Model.from_config(config)


def test_model_valid(model: Model):
    check_model(model)


def test_check_expr_parse_checks_dimensions(model: Model):
    model.dimensions["d1"].str_expr = "sum(co"
    with pytest.raises(ExpressionSyntaxError):
        check_model(model)


def test_check_expr_parse_checks_measures(model: Model):
    model.measures["m1"].str_expr = "sum())))"
    with pytest.raises(ExpressionSyntaxError):
        check_model(model)


def test_check_expr_parse_checks_metrics(model: Model):
    model.metrics["m2"].str_expr = "$$test"
    with pytest.raises(ExpressionSyntaxError):
        check_model(model)


def test_check_expr_parse_checks_filters(model: Model):
    model.measures["m1"].str_filter = "Ahahahah!!!!"
    with pytest.raises(ExpressionSyntaxError):
        check_model(model)


def test_check_expr_parse_checks_table_filters(model: Model):
    model.tables["t1"].filters[0].str_expr = "bleep bloop"
    with pytest.raises(ExpressionSyntaxError):
        check_model(model)


# def test_check_measure_measure_refs(model: Model):
#     model.measures["m1"].str_expr = "$m2"
#     with pytest.raises(InvalidReferenceError):
#         check_model(model)


def test_check_metric_column_refs(model: Model):
    model.metrics["m1"].str_expr = "sum(col)"
    with pytest.raises(InvalidReferenceError):
        check_model(model)


def test_check_metric_dimension_refs(model: Model):
    model.metrics["m1"].str_expr = "sum(:d1)"
    with pytest.raises(InvalidReferenceError):
        check_model(model)


@pytest.mark.xfail
def test_check_measure_filter_measure_refs(model: Model):
    """Not sure if this test is needed.
    # TODO: figure out if referencing metrics in measure filters works the same as
            with dimensions
    """
    model.measures["m1"].str_filter = "$m2"
    with pytest.raises(InvalidReferenceError):
        check_model(model)


def test_check_known_functions_metrics(model: Model):
    model.metrics["m2"].str_expr = "spam($m1 / 100)"
    with pytest.raises(UnknownFunctionError):
        check_model(model)


def test_check_known_functions_dimensions(model: Model):
    model.dimensions["d1"].str_expr = "ham(col)"
    with pytest.raises(UnknownFunctionError):
        check_model(model)


def test_check_known_functions_measures(model: Model):
    model.measures["m1"].str_expr = "sam(whoa)"
    with pytest.raises(UnknownFunctionError):
        check_model(model)


def test_check_known_functions_measure_filter(model: Model):
    model.measures["m1"].str_filter = "gram(gg) > 0"
    with pytest.raises(UnknownFunctionError):
        check_model(model)


def test_check_known_functions_table_filters(model: Model):
    model.tables["t1"].filters[0].str_expr = "fam(lala)"
    with pytest.raises(UnknownFunctionError):
        check_model(model)


def test_check_known_calculations_metrics_metrics(model: Model):
    model.metrics["m1"].str_expr = "$mx"
    with pytest.raises(MissingAggregationError):
        check_model(model)


def test_check_known_calculations_dimensions_dimension(model: Model):
    model.dimensions["d1"].str_expr = ":unk // 10"
    with pytest.raises(MissingDimensionError):
        check_model(model)


def test_check_known_calculations_dimensions_metric(model: Model):
    model.dimensions["d2"].str_expr = "$unk"
    with pytest.raises(MissingAggregationError):
        check_model(model)


def test_check_known_calculations_measure_filters(model: Model):
    model.measures["m1"].str_filter = ":unk > 0"
    with pytest.raises(MissingDimensionError):
        check_model(model)


def test_check_known_calculations_table_filters(model: Model):
    model.tables["t1"].filters[0].str_expr = ":unk > 0"
    with pytest.raises(MissingDimensionError):
        check_model(model)


def test_check_kinds_aggregate_dimension(model: Model):
    model.dimensions["d1"].str_expr = "sum(x) + 1"
    with pytest.raises(AggregateDimensionError):
        check_model(model)


def test_check_kinds_scalar_measure(model: Model):
    model.measures["m1"].str_expr = ":d1 - 'scalar'"
    with pytest.raises(NonAggregateMeasureError):
        check_model(model)


def test_check_kinds_aggregate_measure_filter(model: Model):
    model.measures["m1"].str_filter = "sum(:d1) > 0"
    with pytest.raises(AggregateDimensionError):
        check_model(model)


def test_check_kinds_aggregate_table_filter(model: Model):
    model.tables["t1"].filters[0].str_expr = "count() > 0"
    with pytest.raises(AggregateDimensionError):
        check_model(model)


def test_kinds_mixed_measure(model: Model):
    model.measures["m1"].str_expr = "sum(col) + col"
    with pytest.raises(MixedExpressionError):
        check_model(model)


def test_kinds_mixed_dimension(model: Model):
    model.dimensions["d1"].str_expr = "$m1 + col"
    with pytest.raises(MixedExpressionError):
        check_model(model)


def test_kinds_mixed_measure_filter(model: Model):
    model.measures["m1"].str_filter = "$m1 + col"
    with pytest.raises(MixedExpressionError):
        check_model(model)


def test_kinds_mixed_table_filter(model: Model):
    model.tables["t1"].filters[0].str_expr = "$m1 + col"
    with pytest.raises(MixedExpressionError):
        check_model(model)


def test_dimension_self_ref(model: Model):
    model.dimensions["d1"].str_expr = ":d1 + 1"
    with pytest.raises(CircularReferenceError):
        check_model(model)


def test_metric_self_ref(model: Model):
    model.metrics["m2"].str_expr = "$m2"
    with pytest.raises(CircularReferenceError):
        check_model(model)


def test_check_circular_refs_dimensions(model: Model):
    model.dimensions["d1"].str_expr = ":d2"
    model.dimensions["d2"].str_expr = ":d1"
    with pytest.raises(CircularReferenceError):
        check_model(model)


def test_check_circular_refs_metrics(model: Model):
    model.metrics["m2"].str_expr = "$m3"
    model.metrics["m3"].str_expr = "$m2"
    with pytest.raises(CircularReferenceError):
        check_model(model)


def test_check_measure_references_same_table(model: Model):
    model.measures["me2"].str_expr = "$me1 / count()"
    with pytest.raises(MissingMeasureError):
        check_model(model)


def test_check_measures_use_aggregate_functions(model: Model):
    model.measures["me3"].str_expr = "$me1"
    with pytest.raises(InvalidMeasureError):
        check_model(model)


def test_check_measures_dont_reference_filtered_measures(model: Model):
    model.measures["me1"].str_expr = "$m1 / count()"
    with pytest.raises(FilteredMeasureReferenceError):
        check_model(model)


def test_check_aggregate_dimension_table_has_pk(model: Model):
    model.tables["t1"].primary_key = None
    with pytest.raises(MissingPrimaryKeyError):
        check_model(model)



from itertools import chain

from dictum_core.exceptions import (
    DuplicateCalculationError,
    InvalidIDError,
    MissingCalculationError,
    MissingMetricTableError,
    MissingRelatedKeyError,
    MissingRelatedTableError,
)
from dictum_core.schema.model.model import Model


def _check_id_valid(id: str):
    if not isinstance(id, str):
        raise InvalidIDError(f"Invalid ID: {id}, ID must be a string")
    if not id.isidentifier():
        raise InvalidIDError(f"Invalid ID: {id}, ID must be a valid Python identifier")
    if id.startswith("__"):
        raise InvalidIDError(
            f"Invalid ID {id}, IDs starting from "
            "double underscore '__' are reserved for internal use"
        )
    if id[0].lower() != id[0]:
        raise InvalidIDError(
            f"Invalid ID: {id}, IDs starting with an uppercase letter "
            "are reserved for built-in objects"
        )


def _check_model_ids(model: Model):
    """Check IDs (keys) in the model:
    - tables
    - related tables
    - dimensions
    - metrics
    - unions
    """
    for id_ in model.tables:
        _check_id_valid(id_)
        table = model.tables[id_]
        for id_ in chain(table.related, table.dimensions):
            _check_id_valid(id_)
    for id_ in chain(model.metrics, model.unions):
        _check_id_valid(id_)


def _check_related_tables(model: Model):
    for table_id, table in model.tables.items():
        for alias, related in table.related.items():
            if related.str_table not in model.tables:
                raise MissingRelatedTableError(
                    f"Related table {related.str_table} on table {table_id} "
                    "doesn't exist"
                )
            if (
                related.str_related_key is None
                and model.tables[related.str_table].primary_key is None
            ):
                raise MissingRelatedKeyError(
                    f"Related table {alias} on table {table_id} doesn't "
                    "specify a related key and there no primary key specified for "
                    f"{related.str_table}. "
                    "Either one or another must be specified."
                )


def _check_duplicate_dimensions(model: Model):
    """Check that there are no duplicate dimension names.
    There are no such checks for metrics, tables and unions, because they are defined
    as either dict keys or files, with the file loader checking for duplicates.
    """
    dimensions = {}
    for table_id, table in model.tables.items():
        for key in table.dimensions:
            if key in dimensions:
                raise DuplicateCalculationError(
                    f"Duplicate dimension: {key} on tables "
                    f"{dimensions[key]} and {table_id}"
                )
            dimensions[key] = table_id


def _check_missing_unions(model: Model):
    for table in model.tables.values():
        for dimension_id, dimension in table.dimensions.items():
            if dimension.union is not None:
                # TODO: check that union type and dimension type match
                if dimension.union not in model.unions:
                    raise MissingCalculationError(
                        f"Missing union {dimension.union} is specified "
                        f"on dimension {dimension_id}"
                    )


def _check_metric_table_consistency(model: Model):
    for id_, metric in model.metrics.items():
        if metric.str_time is not None and metric.table is None:
            raise MissingMetricTableError(
                f"Metric {id_} specifies time, but doesn't specify table. "
                "If time is present, table must also be specified."
            )
        if metric.str_filter is not None and metric.table is None:
            raise MissingMetricTableError(
                f"Metric {id_} specifies filter, but doesn't specify table. "
                "If filter is present, table must also be specified."
            )


def _check_missing_metric_time_dimensions(model: Model):
    dimensions = [
        (id_, d) for t in model.tables.values() for id_, d in t.dimensions.items()
    ]
    valid_time_dimension_ids = set(id_ for id_, d in dimensions if "date" in d.type)
    for id_, metric in model.metrics.items():
        if metric.str_time is not None:
            if metric.str_time not in valid_time_dimension_ids:
                raise MissingCalculationError(
                    f"Time dimension {metric.str_time} is specified "
                    f"on metric {id_}, but there's no such dimension in the model."
                )


def _check_metrics_table_exists(model: Model):
    for id_, metric in model.metrics.items():
        if metric.table is not None:
            if metric.table not in model.tables:
                raise MissingMetricTableError(
                    f"Table {metric.table} specified in metric {id_} does not exist"
                )


def check_model_config(model: Model):
    """Check that the model data is valid and internally consistent."""
    _check_model_ids(model)
    _check_related_tables(model)
    _check_duplicate_dimensions(model)
    _check_missing_unions(model)
    _check_metric_table_consistency(model)
    _check_metrics_table_exists(model)
    _check_missing_metric_time_dimensions(model)



from pathlib import Path

import yaml

from dictum_core.examples import chinook
from dictum_core.project.yaml_mapped_dict import YAMLMappedDict


def test_from_path():
    path = Path(chinook.__file__).parent / "metrics"
    result = YAMLMappedDict.from_path(path)
    assert "revenue" in result
    assert isinstance(result["revenue"], YAMLMappedDict)
    assert isinstance(result["revenue"].path, Path)


def test_update_recursive_setitem(tmp_path: Path):
    data = {"a": {"b": {"c": 1, "d": 2}}}
    path = tmp_path / "test.yml"
    path.write_text(yaml.safe_dump(data))
    d = YAMLMappedDict.from_path(path)
    d.update_recursive({"a": {"b": {"c": "test", "e": 3}}})
    d.flush()
    new_data = yaml.safe_load(path.read_text())
    assert new_data == {"a": {"b": {"c": "test", "d": 2, "e": 3}}}

    d["a"]["b"]["e"] = "42"
    d.flush()
    new_data = yaml.safe_load(path.read_text())
    assert new_data == {"a": {"b": {"c": "test", "d": 2, "e": "42"}}}


def test_set_str_to_dict():
    """There was a bug where the recursive update couldn't set a value that
    was previously a str to a dict
    """
    d = YAMLMappedDict({"a": "x"})
    d.update_recursive({"a": {"b": 1}})

    assert d["a"] == {"b": 1}


def test_allow_unicode(tmp_path: Path):
    """Make sure that the YAML produced by dumper has human-readable
    non-ASCII strings
    """
    value = "тест"
    path = tmp_path / "test.yml"
    d = YAMLMappedDict({"a": {"b": value}})
    d.path = path
    d.flush()
    assert value in path.read_text()



import os
from getpass import getpass
from pathlib import Path
from typing import Dict, Optional, Union

import yaml
from jinja2 import Template
from pydantic import BaseModel

from dictum_core.exceptions import MissingPathError
from dictum_core.schema.id import ID


class Env:
    def __init__(self):
        self.data = os.environ.copy()

    def __getattr__(self, key: str):
        try:
            return self.data[key]
        except KeyError:
            val = getpass(key)
            self.data[key] = val
            return val


env = Env()


def _load_yaml_template(path: Path):
    return yaml.load(Template(path.read_text()).render(env=env), Loader=yaml.SafeLoader)


class Profile(BaseModel):
    type: str
    parameters: dict = {}


class Profiles(BaseModel):
    default_profile: str
    profiles: Dict[ID, Profile]


class Project(BaseModel):
    name: str
    description: Optional[str] = None
    locale: str = "en_US"
    currency: str = "USD"

    tables_path: str = "tables"
    metrics_path: str = "metrics"
    unions_path: str = "unions.yml"
    profiles_path: str = "profiles.yml"

    root: Optional[Path] = None

    @classmethod
    def load(cls, path: Union[str, Path]):
        path = Path(path)
        if not path.is_dir():
            raise FileNotFoundError("Project path must be a directory")
        project_yml = path / "project.yml"
        if not project_yml.is_file():
            raise FileNotFoundError(
                "Project must contain a project.yml file at the root"
            )
        data = _load_yaml_template(project_yml)
        data["root"] = str(path)
        return cls.model_validate(data)

    def get_profile(self, profile=None) -> dict:
        profiles_path = self.root / self.profiles_path
        if not profiles_path.exists():
            raise MissingPathError(profiles_path)
        data = _load_yaml_template(self.root / self.profiles_path)
        profiles = Profiles.model_validate(data)
        profile = profiles.default_profile if profile is None else profile
        try:
            return profiles.profiles[profile]
        except KeyError:
            raise KeyError(f"Profile {profile} is not present in profiles")



from itertools import chain

from dictum_core.engine.metrics import transforms as table_transforms
from dictum_core.exceptions import (
    DuplicateColumnError,
    MissingQueryDimensionError,
    MissingQueryMetricError,
    MissingScalarTransformError,
    MissingTableTransformDimensionError,
    MissingTableTransformError,
    ScalarTransformTypeError,
)
from dictum_core.model import Model
from dictum_core.ordered_check_caller import OrderedCheckCaller
from dictum_core.schema.query import Query

check_query = OrderedCheckCaller()


def _check_metrics_exist(model: Model, query: Query):
    for request in query.metrics:
        if request.metric.id not in model.metrics:
            raise MissingQueryMetricError(f"Metric {request.metric.id} does not exist")


def _check_dimension_exists(id_: str, model: Model):
    if id_ not in model.dimensions:
        raise MissingQueryDimensionError(f"Dimension {id_} doesn't exist")


def _check_dimensions_exist(model: Model, query: Query):
    for request in query.dimensions:
        _check_dimension_exists(request.dimension.id, model)
    for dimension in query.filters:
        _check_dimension_exists(dimension.id, model)
    for request in query.metrics:
        for transform in request.metric.transforms:
            for dimension in chain(transform.within, transform.of):
                _check_dimension_exists(dimension.id, model)


@check_query.depends_on(_check_dimensions_exist)
def _check_scalar_transforms_exist(model: Model, query: Query):
    for request in query.dimensions:
        for transform in request.dimension.transforms:
            if transform.id not in model.scalar_transforms:
                raise MissingScalarTransformError(
                    f"Scalar transform {transform.id} does not exist"
                )


@check_query.depends_on(_check_metrics_exist)
def _check_table_transforms_exist(model: Model, query: Query):
    for request in query.metrics:
        for transform in request.metric.transforms:
            if transform.id not in table_transforms:
                raise MissingTableTransformError(
                    f"Table transform {transform.id} does not exist"
                )


@check_query.depends_on(
    _check_dimensions_exist,
    _check_metrics_exist,
)
def _check_dimension_join_paths(model: Model, query: Query):
    query_dimensions = set(
        chain(
            (r.dimension.id for r in query.dimensions),
            (d.id for d in query.filters),
        )
    )
    for request in query.metrics:
        metric = model.metrics[request.metric.id]
        for measure in metric.measures:
            measure_dimensions = set(measure.dimensions)
            missing = query_dimensions - measure_dimensions
            if len(missing) > 0:
                raise MissingQueryDimensionError(
                    f"Dimensions {missing} requested in the query are not available "
                    f"for {measure} in {metric}"
                )


@check_query.depends_on(
    _check_dimensions_exist,
    _check_metrics_exist,
    _check_table_transforms_exist,
    _check_scalar_transforms_exist,
    _check_dimension_join_paths,
)
def _check_of_within_in_dimensions(model: Model, query: Query):
    dimension_digests = set()
    for request in query.dimensions:
        dimension_digests.add(request.dimension.digest)

    for request in query.metrics:
        for transform in request.metric.transforms:
            for dimension in chain(transform.of, transform.within):
                if dimension.digest not in dimension_digests:
                    raise MissingTableTransformDimensionError(
                        f"Dimension request '{dimension.render()}' is used in "
                        f"'{request.metric.render()}', but not in the group by list"
                    )


@check_query.depends_on(_check_scalar_transforms_exist, _check_of_within_in_dimensions)
def _check_scalar_transform_types(model: Model, query: Query):
    for request in query.dimensions:
        dimension = model.dimensions[request.dimension.id]
        for transform in request.dimension.transforms:
            if transform.id not in dimension.transforms:
                raise ScalarTransformTypeError(
                    f"Transform {transform.id} is not available for dimension "
                    f"type {dimension.type.name}"
                )


@check_query.register
def _check_duplicate_columns(model: Model, query: Query):
    names = set()
    for request in chain(query.metrics, query.dimensions):
        if request.name in names:
            raise DuplicateColumnError(
                f"Duplicate column name '{request.name}' in the query"
            )
        names.add(request.name)



import textwrap


class DictumError(Exception):
    """Root exception class"""


class LoadingError(DictumError):
    """Problems occurring when project YAML files are being loaded"""


class MissingPathError(LoadingError):
    """When any of the project paths don't exist"""

    def __init__(self, path, *args) -> None:
        super().__init__(f"Path {path} does not exist", *args)


class DuplicateFileError(LoadingError):
    """When file-based objects stored in nested folders have
    duplicate IDs (filenames)
    """


class ModelError(DictumError):
    """Problems with the model"""


class InvalidIDError(ModelError):
    """When an object's ID is invalid:
    - not a valid Python ID
    - starts with __ (reserved for internal use)
    - starts with an uppercase letter (reserved for internal use)
    """


class CalculationError(ModelError):
    """Errors with calculations (except expressions)"""


class DuplicateCalculationError(CalculationError):
    """When calculations are duplicate, only possible with dimensions"""


class MissingCalculationError(CalculationError):
    """When a calculation referenced outside an expression doesn't exist:
    - union
    - time
    """


class MissingMetricTableError(CalculationError):
    """When a metric must specify a table, but doesn't"""


class TableError(ModelError):
    """Problems with tables"""


class RelatedTableError(TableError):
    """Problems with related tables"""


class MissingRelatedTableError(RelatedTableError):
    """When the specified related table is not present in the model"""


class MissingRelatedKeyError(RelatedTableError):
    """When related_key is not specified and the related table doesn't
    have a primary_key specified
    """


class MissingPrimaryKeyError(TableError):
    """When a table requires a primary_key"""


class ExpressionError(ModelError):
    """Problems with expressions"""


class ExpressionSyntaxError(ExpressionError):
    """Error when parsing with lark"""

    def __init__(self, lark_exc, calc, *args):
        base_text = f"\n{lark_exc.__class__.__name__} error in {calc}: "
        context = textwrap.indent(
            lark_exc.get_context(calc.str_expr),
            prefix=" " * (len(base_text) - 1),
            predicate=lambda x: "^" in x,
        )
        self.lark_exc = lark_exc
        super().__init__(f"{base_text}{context}", *args)


class UnknownFunctionError(ExpressionError):
    """When a function used in the expression doesn't exist"""


class MixedExpressionError(ExpressionError):
    """When an expression mixes aggregate and non-aggregate parts,
    e.g. sum(amount) + tax
    """


class AggregateFunctionArgumentError(ExpressionError):
    """When an aggregate function's argument is aggregate itself"""


class AggregateDimensionError(ExpressionError):
    """When a dimension's expression is aggregate, e.g. sum(amount), except when it's
    a metric reference.
    """


class NonAggregateMeasureError(ExpressionError):
    """When a measure's expression is not aggregate, e.g. `amount * 1000`"""


class ReferenceError(ExpressionError):
    """Problems with expressions referencing other expressions"""


class InvalidReferenceError(ReferenceError):
    """When a certain kind of expression can't reference another kind of expression,
    e.g.
    - measures -> only columns and dimensions
    - metrics -> only other metrics and measures
    - dimensions -> only other dimensions, columns and metrics
    """


class AggregateDimensionPrimaryKeyError(ReferenceError):
    """When there's no primary key on an aggregate dimension's parent table"""


class MissingReferenceError(ReferenceError):
    """Problems with calculations referencing calculations that are not available
    with known joins or don't exist
    """


class MissingDimensionError(MissingReferenceError):
    """When referenced dimension is missing"""


class MissingAggregationError(MissingReferenceError):
    """When referenced metric or measure is missing"""


class MissingMeasureError(MissingReferenceError):
    """When referenced measure on the same table isn't found"""


class FilteredMeasureReferenceError(ReferenceError):
    """When a measure references a filtered measure"""


class InvalidMeasureError(ReferenceError):
    """When a measure doesn't reference any aggregate functions
    and should be a metric
    """


class CircularReferenceError(ReferenceError):
    """When calculations reference each other, and so can't be resolved"""


class ShorthandError(ModelError):
    """Problems with the interactive shorthand syntax for Jupyter"""


class ShorthandSyntaxError(ShorthandError):
    """Nicer syntax errors for shorthands"""

    def __init__(self, lark_exc, definition, *args):
        base_text = f"\n{lark_exc.__class__.__name__} error in expression: "
        context = textwrap.indent(
            lark_exc.get_context(definition),
            prefix=" " * (len(base_text) - 1),
            predicate=lambda x: "^" in x,
        )
        self.lark_exc = lark_exc
        super().__init__(f"{base_text}{context}", *args)


class MissingShorthandTableError(ShorthandError):
    """When a shorthand requires a table, but none is provided"""


class QueryError(DictumError):
    """Problems with the Query object"""


class MissingQueryMetricError(QueryError):
    """When requested metric doesn't exist"""


class MissingQueryDimensionError(QueryError):
    """When requested dimension doesn't exist"""


class MissingScalarTransformError(QueryError):
    """When requested scalar transform doesn't exist"""


class MissingTableTransformError(QueryError):
    """When requested table transform doesn't exist"""


class MissingTableTransformDimensionError(QueryError):
    """When transforms listed in of/within are not present in the query"""


class ScalarTransformTypeError(QueryError):
    """When there's funky business with scalar transform input types"""


class DimensionUnavailableError(QueryError):
    """When dimension exists, but can't be joined"""


class DuplicateColumnError(QueryError):
    """When a query is requesting duplicate columns in the output"""



from datetime import datetime

from dictum_core.format import Format
from dictum_core.model.types import Type
from dictum_core.schema import FormatConfig

locale = "en_US"


def test_format_default_float():
    fmt = Format(locale=locale, type=Type(name="float"))
    assert fmt.format_value(1.0) == "1"
    assert fmt.format_value(None) == ""


def test_format_default_int():
    fmt = Format(locale=locale, type=Type(name="int"))
    assert fmt.format_value(1) == "1"
    assert fmt.format_value(None) == ""


def test_format_default_datetime():
    fmt = Format(locale=locale, type=Type(name="datetime", grain="year"))
    assert fmt.format_value(None) == ""

    dt = datetime(2022, 11, 17, 12, 33, 11)

    def _grain(grain: str):
        return Format(
            locale=locale, type=Type(name="datetime", grain=grain)
        ).format_value(dt)

    assert _grain("year") == "2022"
    assert _grain("quarter") == "Q4 2022"
    assert _grain("month") == "November 2022"
    assert _grain("week") == "week 47 of 2022"
    assert _grain("day") == "11/17/2022"
    assert _grain("hour") == "11/17/22, 12:33\u202fPM"
    assert _grain("minute") == "11/17/22, 12:33\u202fPM"
    assert _grain("second") == "Nov 17, 2022, 12:33:11\u202fPM"


def test_format_default_bool():
    fmt = Format(locale=locale, type=Type(name="bool"))
    assert fmt.format_value(None) == ""
    assert fmt.format_value(True) == "True"


def test_format_default_str():
    fmt = Format(locale=locale, type=Type(name="str"))
    assert fmt.format_value(None) == ""
    assert fmt.format_value("abc") == "abc"


def test_d3_format():
    fmt = Format(
        locale=locale, type=Type(name="float"), config=FormatConfig(kind="percent")
    )
    assert fmt.d3_format == "01,.0%"



from graphlib import TopologicalSorter
from typing import List


class OrderedCheckCaller:
    """Allows the user to specify dependencies for each function
    and then calls all functions with the same arguments, but in the topological
    order.
    """

    def __init__(self) -> None:
        self.graph = {}
        self.fns = {}

    def register(self, fn: callable):
        self.fns[fn.__name__] = fn
        self.graph.setdefault(fn.__name__, set())

    def depends_on(self, *dependencies):
        def decorator(fn):
            self.register(fn)
            for dep in dependencies:
                self.register(dep)
                self.graph[fn.__name__].add(dep.__name__)
            return fn

        return decorator

    def static_order(self) -> List[str]:
        return TopologicalSorter(self.graph).static_order()

    def __call__(self, *args, **kwargs):
        for k in self.static_order():
            self.fns[k](*args, **kwargs)



import importlib
from pathlib import Path
from typing import Optional, Union

import pandas as pd
from lark import Tree

from dictum_core import schema
from dictum_core.backends.base import Backend
from dictum_core.engine import Engine, Result
from dictum_core.exceptions import MissingPathError, MissingShorthandTableError
from dictum_core.model import Model
from dictum_core.project import actions, analyses
from dictum_core.project.calculations import ProjectDimensions, ProjectMetrics
from dictum_core.project.magics import ProjectMagics
from dictum_core.project.magics.parser import (
    parse_shorthand_dimension,
    parse_shorthand_format,
    parse_shorthand_metric,
    parse_shorthand_related,
    parse_shorthand_table,
)
from dictum_core.project.templates import environment
from dictum_core.project.yaml_mapped_dict import YAMLMappedDict
from dictum_core.schema import Query


def _get_subtree_str(text: str, tree: Tree):
    s, e = tree.meta.start_pos, tree.meta.end_pos
    return text[s:e]


def _get_calculation_kwargs(definition: str, tree: Tree) -> dict:
    result = {}

    id_ = next(tree.find_data("id")).children[0]
    result["name"] = id_.replace("_", " ").title()

    expr = next(tree.find_data("expr"))
    result["expr"] = _get_subtree_str(definition, expr)

    for ref in tree.find_data("type"):
        result["type"] = ref.children[0]

    for ref in tree.find_data("table"):
        result["table"] = ref.children[0]

    for ref in tree.find_data("filter"):
        result["filter"] = _get_subtree_str(definition, ref)

    for ref in tree.find_data("alias"):
        result["name"] = ref.children[0]

    for ref in tree.find_data("properties"):
        result.update(ref.children[0])

    return id_, result


class Project:
    def __init__(
        self,
        model_data: YAMLMappedDict,
        backend: Backend,
        project_config: Optional[schema.Project] = None,
    ):
        self.project_config = project_config
        self.backend = backend

        self.model_data = model_data
        model_config = schema.Model.model_validate(model_data)

        self.model = Model.from_config(model_config)

        self.engine = Engine(self.model)
        self.metrics = ProjectMetrics(self)
        self.dimensions = ProjectDimensions(self)
        self.m, self.d = self.metrics, self.dimensions

        self.latest_calc = None

        self.magic()

    @classmethod
    def new(
        cls,
        backend: Backend,
        path: Optional[Path] = None,
        name: str = "Untitled",
        locale: str = "en_US",
        currency: str = "USD",
    ):
        """Create a new project with an empty model. Useful for experimenting
        in Jupyter. If path is provided, creates a new project at that path.
        """
        if path is not None:
            path = Path(path)
            if path.exists() and path.is_dir() and (path / "project.yml").exists():
                print(f"Project already exists, loading project from {path}")
                return Project.from_path(path=path)
            actions.create_new_project(
                path=path, backend=backend, name=name, currency=currency, locale=locale
            )
            print(f"Created a new project at {path}")
            return Project.from_path(path=path)
        model_data = schema.Model(
            name=name, locale=locale, currency=currency
        ).model_dump()
        model_data = YAMLMappedDict(model_data)
        project_config = schema.Project(name=name, locale=locale, currency=currency)
        return cls(
            model_data=model_data, backend=backend, project_config=project_config
        )

    @classmethod
    def from_path(
        cls, path: Optional[Union[str, Path]] = None, profile: Optional[str] = None
    ) -> "Project":
        if path is None:
            path = Path.cwd()
        if isinstance(path, str):
            path = Path(path)
        if not path.exists():
            raise MissingPathError(path)

        project_config = schema.Project.load(path)

        model_data = YAMLMappedDict()
        model_data["name"] = project_config.name
        model_data["description"] = project_config.description
        model_data["locale"] = project_config.locale
        model_data["currency"] = project_config.currency

        tables_path = path / project_config.tables_path
        if not tables_path.is_dir():
            raise MissingPathError(path)
        model_data["tables"] = YAMLMappedDict.from_path(
            path / project_config.tables_path
        )
        model_data["metrics"] = YAMLMappedDict.from_path(
            path / project_config.metrics_path
        )
        model_data["unions"] = YAMLMappedDict.from_path(
            path / project_config.unions_path
        )

        profile = project_config.get_profile(profile)
        backend = Backend.create(profile.type, profile.parameters)

        return cls(
            model_data=model_data, backend=backend, project_config=project_config
        )

    def execute(self, query: Query) -> Result:
        computation = self.engine.get_computation(query)
        return computation.execute(self.backend)

    def query_graph(self, query: Query):
        computation = self.engine.get_computation(query)
        return computation.graph(self.backend)

    def ql(self, query: str) -> analyses.QlQuery:
        return analyses.QlQuery(self, query)

    def select(self, *metrics: str) -> "analyses.Select":
        """
        Select metrics from the project.

        Arguments:
            *metrics: Metric IDs to select.

        Returns:
            A ``Select`` object that can be further modified by chain-calling it's
            methods.
        """
        return analyses.Select(self, *metrics)

    def pivot(self, *metrics: str) -> "analyses.Pivot":
        """Select metrics from the project and construct a pivot table.

        Arguments:
            *metrics: Metric IDs to select.

        Returns:
            A ``Select`` object that can be further modified by chain-calling it's
            methods.
        """
        return analyses.Pivot(self, *metrics)

    @classmethod
    def example(cls, name: str) -> "Project":
        """Load an example project.

        Arguments:
            name (str):
                Name of the example project. Valid values: ``chinook``,
                ``empty``.

        Returns:
            CachedProject: same as ``Project``, but won't read the model config at each
            method invocation.
        """
        example = importlib.import_module(f"dictum_core.examples.{name}.generate")
        result: Project = example.generate()
        # prevent users from changing examples
        result.project_config.root = None
        result.model_data = YAMLMappedDict(result.model_data.dict())
        return result

    def describe(self) -> pd.DataFrame:
        """Show project's metrics and dimensions and their compatibility. If a metric
        can be used with a dimension, there will be a ``+`` sign at the intersection of
        their respective row and column.

        Returns:
            pandas.DataFrame: metric-dimension compatibility matrix
        """
        print(
            f"Project '{self.model.name}', {len(self.model.metrics)} metrics, "
            f"{len(self.model.dimensions)} dimensions. "
            f"Connected to {self.backend}."
        )
        data = []
        for metric in self.model.metrics.values():
            for dimension in metric.dimensions:
                data.append((metric.id, dimension.id, "✚"))
        return (
            pd.DataFrame(data=data, columns=["metric", "dimension", "check"])
            .pivot(index="dimension", columns="metric", values="check")
            .fillna("")
        )

    def magic(self):
        from IPython import get_ipython  # so that linters don't whine

        ip = get_ipython()
        if ip is not None:
            ip.register_magics(ProjectMagics(project=self, shell=ip))
            print(
                r"Magics %ql, %table, %metric, %dimension are registered and bound "
                f"to project {self.project_config.name}"
            )

    def _repr_html_(self):
        template = environment.get_template("project.html.j2")
        return template.render(project=self)

    def update_shorthand_table(self, definition: str):
        tree = parse_shorthand_table(definition)

        table_def, *items = tree.children
        table = table_def.children[0].children[0]
        source = next(table_def.find_data("source"), None)
        if source is not None:
            source = source.children[0]
        if source is None:
            source = table
        data = {"id": table, "source": source}

        pk = next(table_def.find_data("pk"), None)
        if pk is not None:
            data["primary_key"] = pk.children[0]
        self.update_model({"tables": {table: data}})

        # add items
        for item in items:
            if item.data == "related":
                str_shorthand = _get_subtree_str(definition, item)
                self.update_shorthand_related(f"{table}.{str_shorthand}")
            elif item.data == "dimension":
                self.update_shorthand_dimension(
                    _get_subtree_str(definition, item), table
                )
            elif item.data == "metric":
                self.update_shorthand_metric(_get_subtree_str(definition, item), table)
            elif item.data == "table_format":
                self.update_shorthand_format(
                    _get_subtree_str(definition, item.children[0])
                )

    def update_shorthand_related(self, definition: str):
        tree = parse_shorthand_related(definition)
        tables = tree.find_data("table")
        parent = next(tables).children[0]
        target = next(tables, None)
        alias = next(tree.find_data("alias")).children[0]
        if target is None:
            raise MissingShorthandTableError(
                f"\nTable is required for standalone related shorthand: table.{alias}\n"
                "                                                    ^^^^^"
            )
        target = target.children[0]
        # parent, target = list(t.children[0] for t in tree.find_data("table"))
        columns = list(c.children[0] for c in tree.find_data("column"))
        foreign_key = columns[0]
        related_key = None
        if len(columns) == 2:
            related_key = columns[1]
        update = {
            "tables": {
                parent: {
                    "related": {
                        alias: {
                            "table": target,
                            "related_key": related_key,
                            "foreign_key": foreign_key,
                        }
                    }
                }
            }
        }
        self.update_model(update)

    def update_shorthand_metric(self, definition: str, table: Optional[str] = None):
        tree = parse_shorthand_metric(definition)
        id_, kwargs = _get_calculation_kwargs(definition, tree)
        kwargs["table"] = kwargs.get("table", table)
        update = {"metrics": {id_: kwargs}}
        self.update_model(update)
        self.latest_calc = self.model_data["metrics"][id_]

    def update_shorthand_dimension(self, definition: str, table: Optional[str] = None):
        tree = parse_shorthand_dimension(definition)
        id_, kwargs = _get_calculation_kwargs(definition, tree)
        schema.Dimension.model_validate(kwargs)  # validate before updating
        if table is None:
            table = kwargs.pop("table", None)
        if table is None:
            raise ValueError("Table is required, please specify with '@ table'")
        update = {"tables": {table: {"dimensions": {id_: kwargs}}}}
        self.update_model(update)

        self.latest_calc = self.model_data["tables"][table]["dimensions"][id_]

    def update_shorthand_format(self, definition: str):
        format = parse_shorthand_format(definition)
        if isinstance(format.children[0], str):
            format = format.children[0]
        else:
            format = dict(format.children)
        update = {"format": format}
        self.latest_calc.update_recursive(update)
        self.update_model({})  # trigger model update

    def update_model(self, update: dict):
        # avoid updating model_data until model is checked
        new = self.model_data.copy()
        new.update_recursive(update)

        model_config = schema.Model.model_validate(new)
        self.model = Model.from_config(model_config)  # model checks happen here
        # we're ok, can update model data
        self.model_data = new

        # do the rest of the stuff
        self.engine = Engine(self.model)
        self.metrics = ProjectMetrics(self)
        self.dimensions = ProjectDimensions(self)
        self.m, self.d = self.metrics, self.dimensions

    def write(self, path: Optional[Union[str, Path]] = None):
        if isinstance(path, str):
            path = Path(path)

        if path is None and self.project_config.root is None:
            raise ValueError(
                "Project doesn't have a path, please specify as an argument to write()"
            )
        if self.project_config.root is None:
            self.project_config.root = path
        path = self.project_config.root

        if not path.exists():
            actions.create_new_project(
                path,
                backend=self.backend,
                name=self.project_config.name,
                currency=self.project_config.currency,
                locale=self.project_config.locale,
            )

        # assign paths where they're not present
        self.model_data["tables"].assign_paths(
            self.project_config.root / self.project_config.tables_path
        )
        self.model_data["metrics"].assign_paths(
            self.project_config.root / self.project_config.metrics_path
        )
        self.model_data["unions"].assign_paths(
            self.project_config.root / self.project_config.unions_path
        )

        self.model_data.flush()



from pathlib import Path

import yaml
from jinja2 import Template

from dictum_core import project

templates_path = Path(project.__file__).parent / "project_template"


def test_profiles_correct_parameters_indent():
    """There was a bug where only the first parameter was indented correctly with the rest
    indented one level below, e.g.:

    default_profile: ...
    profiles:
      ...:
        type: ...
        parameters:
          param1: ...
        param2: ... <------- incorrect indentation
        param3: ...
    """
    template = Template((templates_path / "profiles.yml").read_text())
    result = template.render(
        profile="test",
        backend="test",
        backend_parameters=yaml.safe_dump({"a": 1, "b": 2, "c": 3}),
    )
    result_data = yaml.safe_load(result)
    assert len(result_data["profiles"]["test"]["parameters"]) == 3



from typing import List, Literal, Optional

from pydantic import BaseModel, Field


class CatalogType(BaseModel):
    name: str
    grain: Optional[str] = None

    class Config:
        orm_mode = True


class CatalogFormat(BaseModel):
    type: CatalogType
    d3_format: Optional[str] = None

    class Config:
        orm_mode = True


class CatalogCalculation(BaseModel):
    id: str
    name: str
    description: Optional[str] = None
    expr: Optional[str] = Field(None, alias="str_expr")
    format: CatalogFormat
    type: CatalogType

    class Config:
        orm_mode = True


class CatalogDimension(CatalogCalculation):
    pass


class MetricLineageItem(BaseModel):
    id: str
    name: str
    type: Literal["Metric", "Measure", "Dimension", "Column"]
    parent: Optional[str] = None


class MetricLineage(BaseModel):
    __root__: List[MetricLineageItem]


class CatalogMetric(CatalogCalculation):
    lineage: MetricLineage
    dimensions: List[CatalogDimension]
    time: List[CatalogDimension] = Field(alias="generic_time_dimensions")



import pytest

from dictum_core.exceptions import (
    DuplicateCalculationError,
    InvalidIDError,
    MissingCalculationError,
    MissingMetricTableError,
    MissingRelatedKeyError,
    MissingRelatedTableError,
)
from dictum_core.schema.model.checks import (
    _check_duplicate_dimensions,
    _check_id_valid,
    _check_metric_table_consistency,
    _check_missing_metric_time_dimensions,
    _check_missing_unions,
    _check_model_ids,
    _check_related_tables,
)
from dictum_core.schema.model.model import Model


def test_check_duplicate_dimensions():
    model = Model.model_validate(
        {
            "name": "test",
            "tables": {
                "t1": {
                    "source": "t1",
                    "dimensions": {
                        "dupe": {"name": "dupe", "expr": "x", "type": "int"}
                    },
                },
                "t2": {
                    "source": "t2",
                    "dimensions": {
                        "dupe": {"name": "dupe", "expr": "x", "type": "int"}
                    },
                },
            },
        }
    )
    with pytest.raises(DuplicateCalculationError):
        _check_duplicate_dimensions(model)


def test_check_model_ids_checks_tables():
    model = Model.model_validate(
        {
            "name": "test",
            "tables": {"__t1": {"source": "t1"}},
        }
    )
    with pytest.raises(InvalidIDError):
        _check_model_ids(model)


def test_check_model_ids_checks_metrics():
    model = Model.model_validate(
        {"name": "test", "metrics": {"__m1": {"name": "m1", "expr": "m1"}}}
    )
    with pytest.raises(InvalidIDError):
        _check_model_ids(model)


def test_check_model_ids_checks_unions():
    model = Model.model_validate(
        {"name": "test", "unions": {"__id": {"name": "test", "type": "test"}}}
    )
    with pytest.raises(InvalidIDError):
        _check_model_ids(model)


def test_check_model_ids_checks_related():
    model = Model.model_validate(
        {
            "name": "test",
            "tables": {
                "test": {
                    "source": "test",
                    "related": {"__id": {"table": "test", "foreign_key": "key"}},
                }
            },
        }
    )
    with pytest.raises(InvalidIDError):
        _check_model_ids(model)


def test_check_model_ids_checks_dimensions():
    model = Model.model_validate(
        {
            "name": "test",
            "tables": {
                "t1": {
                    "source": "src",
                    "dimensions": {
                        "__id": {
                            "name": "test",
                            "expr": "test",
                            "type": "test",
                        }
                    },
                }
            },
        }
    )
    with pytest.raises(InvalidIDError):
        _check_model_ids(model)


def test_check_id_valid_checks_dunder():
    with pytest.raises(InvalidIDError):
        _check_id_valid("__id")


def test_check_id_valid_checks_uppercase_first_character():
    with pytest.raises(InvalidIDError):
        _check_id_valid("Test")


def test_check_id_valid_checks_identifier():
    with pytest.raises(InvalidIDError):
        _check_id_valid("not id")


def test_check_id_valid_ok():
    _check_id_valid("va1id_id")


def test_check_metric_table_consistency_checks_time():
    model = Model.model_validate(
        {
            "name": "test",
            "metrics": {
                "test": {
                    "name": "test",
                    "expr": "test",
                    "time": "test",
                }
            },
        }
    )
    with pytest.raises(MissingMetricTableError):
        _check_metric_table_consistency(model)


def test_check_metric_table_consistency_checks_filter():
    model = Model.model_validate(
        {
            "name": "test",
            "metrics": {
                "test": {
                    "name": "test",
                    "expr": "test",
                    "filter": "test",
                }
            },
        }
    )
    with pytest.raises(MissingMetricTableError):
        _check_metric_table_consistency(model)


def test_check_missing_time_dimensions_checks_metrics():
    model = Model.model_validate(
        {
            "name": "test",
            "metrics": {
                "test": {
                    "name": "test",
                    "expr": "test",
                    "table": "test",
                    "time": "test",
                }
            },
        }
    )
    with pytest.raises(MissingCalculationError):
        _check_missing_metric_time_dimensions(model)


def test_check_missing_unions():
    model = Model.model_validate(
        {
            "name": "test",
            "tables": {
                "t1": {
                    "source": "test",
                    "dimensions": {
                        "test": {
                            "name": "test",
                            "type": "test",
                            "expr": "test",
                            "union": "missing",
                        }
                    },
                }
            },
        }
    )
    with pytest.raises(MissingCalculationError):
        _check_missing_unions(model)


def test_check_related_tables_checks_table():
    model = Model.model_validate(
        {
            "name": "test",
            "tables": {
                "t1": {
                    "source": "test",
                    "related": {"rel": {"table": "missing", "foreign_key": "fk"}},
                }
            },
        }
    )
    with pytest.raises(MissingRelatedTableError):
        _check_related_tables(model)


def check_test_related_tables_checks_related_key():
    model = Model.model_validate(
        {
            "name": "test",
            "tables": {
                "t1": {
                    "source": "test",
                    "related": {"t2": {"table": "t2", "foreign_key": "fk"}},
                },
                "t2": {"source": "test"},
            },
        }
    )
    with pytest.raises(MissingRelatedKeyError):
        _check_related_tables(model)



from dictum_core.model import Model


def test_measure_dimensions_union(chinook: Model):
    assert "country" in set(chinook.measures.get("n_customers").dimensions)
    assert "country" in set(d.id for d in chinook.metrics.get("n_customers").dimensions)


def test_metric_dimensions(chinook: Model):
    dimensions = chinook.metrics["revenue_per_track"].dimensions
    assert isinstance(dimensions, list)
    assert len(dimensions) == 7


def test_metric_generic_time_dimensions(chinook: Model):
    assert chinook.metrics["revenue_per_track"].generic_time_dimensions == []
    assert len(chinook.metrics["avg_sold_unit_price"].generic_time_dimensions) == 7
    assert (
        chinook.metrics["avg_sold_unit_price"].generic_time_dimensions[1].id == "Year"
    )
    assert (
        chinook.metrics["avg_sold_unit_price"].generic_time_dimensions[1].sort_order
        == 1
    )



from collections import UserDict
from collections.abc import MutableMapping
from itertools import chain
from pathlib import Path

import yaml

from dictum_core.exceptions import DuplicateFileError, MissingPathError


def _update_recursive(d, u):
    for k, v in u.items():
        if isinstance(v, dict) and isinstance(d.get(k, {}), MutableMapping):
            d[k] = _update_recursive(d.get(k, {}), v)
        else:
            d[k] = v
    return d


class YAMLMappedDict(UserDict):
    file_extensions = {".yaml", ".yml"}

    def __init__(self, data=None, /, **kwargs):
        if data is None:
            data = {}
        self.path = None
        self.parent = None

        for k, v in data.items():
            if isinstance(v, dict):
                data[k] = YAMLMappedDict(v)
                data[k].parent = self
        return super().__init__(data, **kwargs)

    @classmethod
    def from_path(cls, path: Path):
        if not path.exists():
            raise MissingPathError(path)
        if path.is_file() and path.suffix in cls.file_extensions:
            value = yaml.safe_load(path.read_text())
            result = cls(value)
            result.path = path
            return result
        if path.is_dir():
            items = {}
            ids = set()
            for subpath in chain(
                *(path.glob(f"**/*{ext}") for ext in cls.file_extensions)
            ):
                key = subpath.stem
                if key in ids:
                    raise DuplicateFileError(
                        f"Duplicate filenames at {path}: {subpath.name}"
                    )
                ids.add(key)
                items[key] = cls.from_path(subpath)
            return cls(items)

    def flush(self):
        if self.path is not None:
            self.path.write_text(
                yaml.safe_dump(self.dict(), sort_keys=False, allow_unicode=True)
            )
        for v in self.data.values():
            if isinstance(v, YAMLMappedDict):
                v.flush()

    def __setitem__(self, key: str, item):
        if isinstance(item, dict):
            item = YAMLMappedDict(item)
            item.parent = self
        self.data[key] = item

    def update_recursive(self, update: dict):
        _update_recursive(self, update)

    def dict(self) -> dict:
        result = {}
        for k, v in self.data.items():
            if isinstance(v, YAMLMappedDict):
                result[k] = v.dict()
            else:
                result[k] = v
        return result

    def assign_paths(self, base_path: Path):
        if base_path.is_dir():
            for k, v in self.items():
                if v.path is None:
                    v.path = base_path / f"{k}.yml"
        elif base_path.is_file():
            if self.path is None:
                self.path = base_path
        else:
            raise FileNotFoundError(f"{base_path} does not exist")

    def copy(self) -> "YAMLMappedDict":
        return YAMLMappedDict(self.dict())



from pathlib import Path

import pytest
import yaml
from lark import Tree

from dictum_core.exceptions import ShorthandSyntaxError
from dictum_core.project.magics.magics import ProjectMagics
from dictum_core.project.magics.parser import (
    parse_shorthand_dimension,
    parse_shorthand_format,
    parse_shorthand_metric,
    parse_shorthand_table,
)
from dictum_core.project.project import Project


def test_parse_raw_table():
    parse_shorthand_table("test") == Tree("table_full", [Tree("table", ["test"])])


def test_parse_table_with_pk():
    parse_shorthand_table("test[pk]") == Tree(
        "table_full", [Tree("table", ["test"]), Tree("pk", ["pk"])]
    )


def test_parse_table_with_pk_source():
    result = parse_shorthand_table("test[pk] src")
    assert result == Tree(
        "table_full",
        [
            Tree(
                "table_def",
                [
                    Tree("table", ["test"]),
                    Tree("pk", ["pk"]),
                    Tree("source", ["src"]),
                ],
            )
        ],
    )


def test_parse_table_with_pk_source_kv():
    result = parse_shorthand_table("test[pk] { src=x y=z }")
    assert result == Tree(
        "table_full",
        [
            Tree(
                "table_def",
                [
                    Tree("table", ["test"]),
                    Tree("pk", ["pk"]),
                    Tree("source", [{"src": "x", "y": "z"}]),
                ],
            )
        ],
    )


def test_parse_table_with_source():
    parse_shorthand_table("test { schema=public table=what }").children == [
        Tree("table", ["test"]),
        Tree("source", [{"schema": "public", "table": "what"}]),
    ]


@pytest.mark.parametrize("tbl", ["", " tbl. "])
@pytest.mark.parametrize("type", ["", " :: int "])
@pytest.mark.parametrize("alias", ["", " as z "])
@pytest.mark.parametrize("filter", ["", "where :dim = 'val'"])
@pytest.mark.parametrize("props", ["", "{ k=v a=b }"])
def test_parse_metric(tbl, type, alias, filter, props):
    """Basic tests for various combinations of metric parameters"""
    definition = f"{tbl} x = sum(y) {filter} {type} {alias} {props}"
    parse_shorthand_metric(definition)


@pytest.mark.parametrize("alias", ["", " as z "])
@pytest.mark.parametrize("type", [" :: int "])
@pytest.mark.parametrize("tbl", ["", " tbl. "])
@pytest.mark.parametrize("props", ["", "{ k=v a=b }"])
def test_parse_dimension(tbl, type, alias, props):
    """Basic tests for various combinations of dimension parameters"""
    definition = f"{tbl} x = y {type} {alias} {props}"
    parse_shorthand_dimension(definition)


def test_dimension_without_type_fails():
    with pytest.raises(ShorthandSyntaxError):
        parse_shorthand_dimension("x = y")


def test_parse_table_with_related():
    result = parse_shorthand_table("table related alias = column -> other.column")
    _, related = result.children
    assert related.data == "related"


def test_parse_format_str():
    result = parse_shorthand_format("currency")
    assert result == Tree("format", ["currency"])


def test_parse_format_kv():
    result = parse_shorthand_format("kind=currency currency=USD")
    assert result == Tree("format", [("kind", "currency"), ("currency", "USD")])


def test_standalone_table(tmp_path: Path, project):
    project = Project.new(backend=project.backend, path=tmp_path)
    magics = ProjectMagics(project)
    magics.table("genres")


@pytest.fixture(scope="function")
def empty():
    return Project.example("empty")


def test_project_create_table(empty: Project):
    empty.update_shorthand_table("invoice_items")
    assert "invoice_items" in empty.model_data["tables"]


def test_project_create_metric_with_filter(empty: Project):
    empty.update_shorthand_table("tbl")
    empty.update_shorthand_metric("tbl.x = sum(x * y) where z > 0")
    assert "x" in empty.model_data["metrics"]
    assert "filter" in empty.model_data["metrics"]["x"]
    assert empty.model_data["metrics"]["x"]["filter"] == "z > 0"


def test_metric_properties(empty: Project):
    empty.update_shorthand_table("tbl")
    empty.update_shorthand_dimension("tbl.d = d ::date")
    empty.update_shorthand_metric(
        'tbl.x = sum(x * y) { time=d name="something something" }'
    )
    assert "x" in empty.model.metrics
    assert empty.model.metrics["x"].name == "something something"
    assert len(empty.model.metrics["x"].generic_time_dimensions) > 0


def test_project_create_from_scratch_write(tmp_path: Path, project: Project):
    project = Project.new(backend=project.backend, path=tmp_path)
    magics = ProjectMagics(project)
    magics.table("invoice_items")
    magics.metric(
        "invoice_items.revenue = sum(Quantity * UnitPrice) { format=currency }"
    )
    project.write()
    assert (
        yaml.safe_load((tmp_path / "metrics" / "revenue.yml").read_text())["format"]
        == "currency"
    )
    magics.table("invoices")
    magics.related("invoice_items.invoice = InvoiceId -> invoices.InvoiceId")
    magics.dimension("invoices.date = InvoiceDate ::date")
    magics.table("genres")
    magics.dimension("genres.genre = Name ::str")
    cell = """
    dimension media_type = Name ::str
    dimension music = MediaTypeId in (1, 2, 4, 5) ::bool as "Is Music"
    """
    magics.table(line="media_types", cell=cell)
    cell = """
    related genre = GenreId -> genres.GenreId
    related media_type = MediaTypeId -> media_types.MediaTypeId
    metric tracks = count()
    dimension track_length = Milliseconds / 1000 / 60 ::float
    """
    magics.table(line="tracks[TrackId]", cell=cell)
    magics.related("invoice_items.track = TrackId -> tracks")
    magics.metric("invoice_items.unique_paying_customers = countd(invoice.CustomerId)")
    magics.metric("arppu = $revenue / $unique_paying_customers as ARPPU")
    project.write()

    assert (tmp_path / "project.yml").is_file()
    assert (tmp_path / "metrics" / "arppu.yml").is_file()

    # edit
    magics.metric(
        "arppu = $revenue / $unique_paying_customers "
        'as "Average Revenue per Paying User"'
    )
    project.write()
    arppu = yaml.safe_load((tmp_path / "metrics" / "arppu.yml").read_text())
    assert arppu["name"] == "Average Revenue per Paying User"



from dictum_core.schema.model.calculations import Calculation


def test_expr_alias():
    calc = Calculation.model_validate(
        {"id": "test", "name": "test", "type": "int", "expr": "str_expr"}
    )
    assert calc.str_expr == "str_expr"



from typing import Dict, List

from dictum_core.format import Format


class Formatter:
    """Intelligently format the data based on the information in the query."""

    def __init__(self, formats: Dict[str, Format]):
        self.formats = formats

    def format(self, data: List[dict]):
        for row in data:
            yield self.format_row(row)

    def format_row(self, row: dict):
        return {k: self.format_value(v, self.formats[k]) for k, v in row.items()}

    def format_value(self, value, format: Format) -> str:
        return format.format_value(value)



from typing import List, Optional

from pydantic import BaseModel

from dictum_core.schema.catalog.calculations import CatalogMetric


class Catalog(BaseModel):
    name: str
    description: Optional[str] = None
    metrics: List[CatalogMetric]



from datetime import datetime
from functools import lru_cache
from numbers import Number
from typing import Any, Optional

from babel import Locale
from babel.dates import format_datetime, match_skeleton
from babel.numbers import format_currency, format_decimal, format_percent

from dictum_core.format import d3
from dictum_core.model.types import Type
from dictum_core.schema.model.format import Format, FormatConfig

grain_skeletons = {
    "year": "y",
    "quarter": "yQQQ",
    "month": "MMMMy",
    "week": "yw",
    "day": "yMd",
}

cache = lru_cache(maxsize=None)


@cache
def get_locale(locale: str) -> Locale:
    return Locale(locale)


class BaseFormatKind:
    def __init__(self, locale: str):
        self.locale = locale

    @property
    def localedata(self) -> Locale:
        return get_locale(self.locale)

    def format_value(self, value: Any) -> str:
        raise NotImplementedError

    def get_d3_format(self) -> str:
        return None


class StringFormat(BaseFormatKind):
    def __init__(self, locale: str):
        super().__init__(locale)

    def format_value(self, value: str) -> str:
        return value


class BoolFormat(BaseFormatKind):
    def __init__(self, locale: str):
        super().__init__(locale)

    def format_value(self, value: bool) -> str:
        return str(value)  # TODO: format based on locale


class NumberFormat(BaseFormatKind):
    def __init__(self, locale: str, pattern: Optional[str] = None):
        super().__init__(locale)
        self.pattern = pattern

    def format_value(self, value: Number) -> str:
        return super().format_value(value)


class DecimalFormat(NumberFormat):
    def format_value(self, value: Number) -> str:
        return format_decimal(value, format=self.pattern, locale=self.locale)

    def get_d3_format(self) -> str:
        pattern = self.pattern
        if pattern is None:
            pattern = self.localedata.decimal_formats[None].pattern
        return d3.ldml_number_to_d3_format(pattern)


class PercentFormat(NumberFormat):
    def format_value(self, value: Number) -> str:
        if self.pattern is None:
            return format_percent(value, locale=self.locale)
        return format_decimal(value, format=self.pattern, locale=self.locale)

    def get_d3_format(self) -> str:
        pattern = self.pattern
        if pattern is None:
            pattern = self.localedata.percent_formats[None].pattern
        return d3.ldml_number_to_d3_format(pattern)


class CurrencyFormat(NumberFormat):
    def __init__(self, locale: str, currency: str, pattern: Optional[str] = None):
        super().__init__(locale, pattern)
        self.currency = currency

    def format_value(self, value: Number) -> str:
        return format_currency(
            value, currency=self.currency, format=self.pattern, locale=self.locale
        )

    def get_d3_format(self) -> str:
        pattern = self.pattern
        if pattern is None:
            pattern = self.localedata.currency_formats["standard"].pattern
        return d3.ldml_number_to_d3_format(pattern)


class DatetimeFormat(BaseFormatKind):
    def __init__(
        self,
        locale: str,
        pattern: Optional[str] = None,
        skeleton: Optional[str] = None,
        grain: Optional[str] = None,
    ):
        super().__init__(locale)
        self.pattern = pattern
        self.skeleton = skeleton
        self.grain = grain

    def get_skeleton_pattern(self, skeleton: str):
        key = match_skeleton(skeleton, self.localedata.datetime_skeletons)
        if key is None:
            return None
        return self.localedata.datetime_skeletons[key]

    def get_babel_format(self):
        if self.pattern is not None:
            return self.pattern

        if self.skeleton is not None:
            return self.get_skeleton_pattern(self.skeleton)

        grain_skeleton = grain_skeletons.get(self.grain)
        if grain_skeleton is not None:
            return self.get_skeleton_pattern(grain_skeleton)

        format = "short"
        if self.grain == "second":
            format = "medium"

        date_format = self.localedata.date_formats[format]
        time_format = self.localedata.time_formats[format]
        return self.localedata.datetime_formats[format].format(time_format, date_format)

    def format_value(self, value: datetime) -> str:
        return format_datetime(
            value, format=self.get_babel_format(), locale=self.locale
        )

    def get_d3_format(self) -> str:
        pattern = self.get_babel_format()
        return d3.ldml_date_to_d3_time_format(pattern)


class Format:
    def __init__(
        self,
        locale: str,
        type: Type,
        default_currency: Optional[str] = None,
        config: Optional[Format] = None,
    ):
        self.locale = locale
        self.type = type
        self.default_currency = default_currency
        if isinstance(config, str):
            config = FormatConfig(kind=config)
        if config is not None and config.kind == "currency" and config.currency is None:
            config.currency = default_currency
        self.config = config

    @property
    def kind(self) -> BaseFormatKind:
        if not isinstance(self.config, FormatConfig):
            if self.type.name == "str":
                return StringFormat(locale=self.locale)
            if self.type.name == "bool":
                return BoolFormat(locale=self.locale)
            if self.type.name in {"int", "float"}:
                return DecimalFormat(locale=self.locale)
            if self.type.name == "datetime":
                return DatetimeFormat(locale=self.locale, grain=self.type.grain)

        if self.config.kind == "string":
            return StringFormat(locale=self.locale)
        if self.config.kind in {"number", "decimal"}:
            return DecimalFormat(locale=self.locale, pattern=self.config.pattern)
        if self.config.kind == "currency":
            return CurrencyFormat(
                locale=self.locale,
                currency=self.config.currency,
                pattern=self.config.pattern,
            )
        if self.config.kind == "percent":
            return PercentFormat(locale=self.locale, pattern=self.config.pattern)
        if self.config.kind in {"datetime", "date"}:
            return DatetimeFormat(
                locale=self.locale,
                pattern=self.config.pattern,
                skeleton=self.config.skeleton,
                grain=self.type.grain,
            )

    @property
    def currency(self) -> str:
        if self.config is None:
            return None
        return self.config.currency

    @property
    def d3_format(self) -> Optional[str]:
        return self.kind.get_d3_format()

    def format_value(self, value: Any) -> str:
        if value is None:
            return ""
        return self.kind.format_value(value)



from unittest import mock

import pytest

from dictum_core.backends.base import ExpressionTransformer
from dictum_core.backends.secret import Secret
from dictum_core.backends.sql_alchemy import (
    ColumnTransformer,
    SQLAlchemyBackend,
    SQLAlchemyCompiler,
)
from dictum_core.project.actions import _get_backend_parameters


@pytest.fixture(scope="module", autouse=True)
def patch_sqlalchemy_url():
    with mock.patch(
        "dictum_core.backends.sql_alchemy.SQLAlchemyBackend.url", "sqlite://"
    ):
        yield


def test_compiler_case():
    compiler = mock.MagicMock()
    transformer = ExpressionTransformer(compiler)
    transformer.case([1, 2, 3, 4])
    compiler.case.assert_called_once_with([[1, 2], [3, 4]], else_=None)

    transformer.case([1, 2, 3, 4, 5])
    compiler.case.assert_called_with([[1, 2], [3, 4]], else_=5)


def test_sqlalchemy_no_default_schema():
    backend = SQLAlchemyBackend()
    with mock.patch("dictum_core.backends.sql_alchemy.Table") as Table:
        backend.table("table")
        assert Table.call_args[1]["schema"] is None


def test_sqlalchemy_default_schema():
    backend = SQLAlchemyBackend(default_schema="test")
    with mock.patch("dictum_core.backends.sql_alchemy.Table") as Table:
        backend.table("table")
        assert Table.call_args[1]["schema"] == "test"


def test_sqlalchemy_override_default_schema():
    backend = SQLAlchemyBackend(default_schema="test")
    with mock.patch("dictum_core.backends.sql_alchemy.Table") as Table:
        backend.table("table", schema="override")
        assert Table.call_args[1]["schema"] == "override"


def test_secret_parameters():
    class SecretBackend(SQLAlchemyBackend):
        type = "secret"

        def __init__(self, secret: Secret = "secret", x: int = 1):
            super().__init__(secret=secret, x=x)

    backend = SecretBackend()
    parameters = _get_backend_parameters(backend)
    assert parameters["x"] == 1
    assert parameters["secret"] == "{{ env.DICTUM_SECRET_SECRET_SECRET }}"


def test_sqlalchemy_case():
    import sqlalchemy as sa

    from dictum_core.model.expr import parse_expr

    parsed = parse_expr("count(case when tbl.type = 1 then 1 end) / count()")
    backend = SQLAlchemyBackend()
    compiler = SQLAlchemyCompiler(backend)
    ct = ColumnTransformer(
        {
            "tbl": sa.Table("tbl", sa.MetaData(), sa.Column("type", sa.Integer)),
        }
    )
    columned = ct.transform(parsed)
    compiler.transformer.transform(columned)



import pytest

from dictum_core import Project
from dictum_core.engine.checks import check_query
from dictum_core.exceptions import (
    DuplicateColumnError,
    MissingQueryDimensionError,
    MissingQueryMetricError,
    MissingScalarTransformError,
    MissingTableTransformDimensionError,
    MissingTableTransformError,
    ScalarTransformTypeError,
)
from dictum_core.model import Model
from dictum_core.schema import Query


@pytest.fixture(scope="function")
def query(project: Project):
    """Provides a valid query that can be broken for tests"""
    return project.ql(
        """
    select revenue, revenue.percent within (invoice_date.year)
    where genre = 'Rock'
    by invoice_date.year, invoice_date.quarter
    """
    ).query


@pytest.fixture(scope="module")
def model(project: Project):
    return project.model


def test_query_valid(model: Model, query: Query):
    check_query(model, query)


def test_check_metrics_exist(model: Model, query: Query):
    query.metrics[0].metric.id = "missing"
    with pytest.raises(MissingQueryMetricError):
        check_query(model, query)


def test_check_dimensions_exist(model: Model, query: Query):
    query.dimensions[0].dimension.id = "missing"
    with pytest.raises(MissingQueryDimensionError):
        check_query(model, query)


def test_check_dimensions_exist_checks_filters(model: Model, query: Query):
    query.filters[0].id = "missing"
    with pytest.raises(MissingQueryDimensionError):
        check_query(model, query)


def test_check_dimensions_exist_checks_of_within(model: Model, query: Query):
    query.metrics[1].metric.transforms[0].within[0].id = "missing"
    with pytest.raises(MissingQueryDimensionError):
        check_query(model, query)


def test_check_scalar_transforms_exist(model: Model, query: Query):
    query.dimensions[0].dimension.transforms[0].id = "missing"
    with pytest.raises(MissingScalarTransformError):
        check_query(model, query)


def test_check_table_transforms_exist(model: Model, query: Query):
    query.metrics[1].metric.transforms[0].id = "missing"
    with pytest.raises(MissingTableTransformError):
        check_query(model, query)


def test_check_of_within_in_dimensions(model: Model, query: Query):
    query.dimensions = []
    with pytest.raises(MissingTableTransformDimensionError):
        check_query(model, query)


def test_check_scalar_transform_types(model: Model, query: Query):
    dimension = query.dimensions[0].dimension
    dimension.transforms[0].id = "step"
    dimension.transforms[0].args = [10]
    query.dimensions[0].dimension = dimension
    query.metrics[1].metric.transforms[0].within[0] = dimension
    with pytest.raises(ScalarTransformTypeError):
        check_query(model, query)


def test_check_dimension_join_paths(model: Model):
    query = Query.model_validate(
        {
            "metrics": [{"metric": {"id": "track_count"}}],
            "dimensions": [{"dimension": {"id": "invoice_date"}}],
        }
    )
    with pytest.raises(MissingQueryDimensionError, match="not available for Measure"):
        check_query(model, query)


def test_check_duplicate_columns(model: Model, query: Query):
    query.dimensions[0].alias = "revenue"
    with pytest.raises(DuplicateColumnError):
        check_query(model, query)



from typing import Literal, Optional, Union

from pydantic import BaseModel, root_validator

FormatKind = Literal[
    "number", "decimal", "percent", "currency", "date", "datetime", "string"
]


class FormatConfig(BaseModel):
    kind: FormatKind
    pattern: Optional[str] = None
    skeleton: Optional[str] = None
    currency: Optional[str] = None

    @root_validator(skip_on_failure=True)
    def validate_pattern_skeleton(cls, values):
        pat = values.get("pattern")
        skel = values.get("skeleton")
        if pat is not None and skel is not None:
            raise ValueError("pattern and skeleton options are mutually exclusive")
        if skel is not None and values["kind"] not in {"date", "datetime"}:
            raise ValueError(
                "skeletons can only be used with date and datetime formats"
            )
        return values


Format = Union[FormatKind, FormatConfig]


class Formatted(BaseModel):
    format: Optional[Format] = None



from itertools import chain

from lark.exceptions import UnexpectedInput

from dictum_core.exceptions import (
    AggregateDimensionError,
    CircularReferenceError,
    ExpressionSyntaxError,
    FilteredMeasureReferenceError,
    InvalidMeasureError,
    InvalidReferenceError,
    MissingAggregationError,
    MissingDimensionError,
    MissingMeasureError,
    MissingPrimaryKeyError,
    NonAggregateMeasureError,
    UnknownFunctionError,
)
from dictum_core.model.calculations import Expression, Measure
from dictum_core.model.expr.introspection import (
    ExprKind,
    aggregate_functions,
    get_expr_kind,
)
from dictum_core.model.model import Model
from dictum_core.ordered_check_caller import OrderedCheckCaller

check_model = OrderedCheckCaller()


def _check_expr_syntax(expression: Expression):
    try:
        expression.parsed_expr
    except UnexpectedInput as e:
        raise ExpressionSyntaxError(e, expression)


def _check_expr_parse(model: Model):
    for expression in model.expressions:
        _check_expr_syntax(expression)


@check_model.depends_on(_check_expr_parse)
def _check_metrics_dont_use_dimension_refs(model: Model):
    for metric in model.metrics.values():
        for ref in metric.parsed_expr.find_data("dimension"):
            id_ = ref.children[0]
            raise InvalidReferenceError(
                f"{metric} references :{id_}, "
                "dimension references in metrics are not allowed. "
                "Consider specifying table."
            )


@check_model.depends_on(_check_expr_parse)
def _check_metrics_dont_use_column_refs(model: Model):
    for metric in model.metrics.values():
        for ref in metric.parsed_expr.find_data("column"):
            id_ = ".".join(ref.children)
            raise InvalidReferenceError(
                f"{metric} references column {id_}, "
                "column references in metrics are not allowed. "
                "Consider specifying table."
            )


known_functions = {
    "sum",
    "count",
    "countd",
    "min",
    "max",
    "avg",
    "abs",
    "floor",
    "ceil",
    "coalesce",
    "tointeger",
    "tofloat",
    "todate",
    "todatetime",
    "now",
    "today",
    "datediff",
    "datepart",
    "datetrunc",
    "IF",
}


def _check_expression_functions(expression: Expression):
    for ref in expression.parsed_expr.find_data("call"):
        fn, *_ = ref.children
        if fn not in known_functions:
            raise UnknownFunctionError(f"unknown function: {fn}")


@check_model.depends_on(_check_expr_parse)
def _check_known_functions(model: Model):
    for expression in model.expressions:
        _check_expression_functions(expression)


def _check_expression_known_calculations(model: Model, expression: Expression):
    for ref in expression.parsed_expr.find_data("dimension"):
        id_ = ref.children[0]
        if id_ not in model.dimensions:
            raise MissingDimensionError(
                f"{expression} references an unknown dimension: {id_}"
            )
    for ref in expression.parsed_expr.find_data("measure"):
        id_ = ref.children[0]
        if id_ not in set(chain(model.measures, model.metrics)):
            raise MissingAggregationError(
                f"{expression} references an unknown metric or measure: {id_}"
            )


@check_model.depends_on(_check_expr_parse)
def _check_known_calculations(model: Model):
    for expression in model.expressions:
        _check_expression_known_calculations(model, expression)


@check_model.depends_on(_check_expr_parse, _check_known_calculations)
def _check_measure_references_same_table(model: Model):
    for measure in model.measures.values():
        for ref in measure.parsed_expr.find_data("measure"):
            id_ = ref.children[0]
            if id_ not in measure.table.measures:
                raise MissingMeasureError(
                    f"{measure} references {id_}, "
                    f"but it's not present on {measure.table}"
                )


def _check_no_aggregate_functions(expression: Expression):
    for ref in expression.parsed_expr.find_data("call"):
        fn, *_ = ref.children
        if fn in aggregate_functions:
            raise AggregateDimensionError(
                f"Aggregate function {fn} in {expression} "
                f"expression: '{expression.str_expr}'"
            )


@check_model.depends_on(_check_expr_parse, _check_known_functions)
def _check_calculation_kinds(model: Model):
    # dimensions can be aggregate, so no direct check
    # validate against using aggregate functions in dimensions
    for dimension in model.dimensions.values():
        if not isinstance(dimension, Expression):
            continue
        _check_no_aggregate_functions(dimension)
        get_expr_kind(dimension.parsed_expr)

    for measure in model.measures.values():
        kind = get_expr_kind(measure.parsed_expr)
        if kind != ExprKind.aggregate:
            raise NonAggregateMeasureError(
                f"{measure} expression '{measure.str_expr}' is not aggregate"
            )
        if measure.filter is not None:
            _check_no_aggregate_functions(measure.filter)
            get_expr_kind(measure.filter.parsed_expr)  # check for mixed

    for table in model.tables.values():
        for table_filter in table.filters:
            _check_no_aggregate_functions(table_filter)
            get_expr_kind(table_filter.parsed_expr)  # check for mixed


def _check_expression_circular_refs(model: Model, expression: Expression, path=None):
    if path is None:
        path = ()
    for ref in expression.parsed_expr.find_data("measure"):
        id_ = ref.children[0]

        # metrics can self-reference by referencing a same-named measure
        if id_ in model.measures:
            continue

        if id_ in path:
            path = " -> ".join(path + (id_,))
            raise CircularReferenceError(f"Circular reference in calculations: {path}")

        # measures can't reference stuff, so we don't check for that
        # metric refs existence is checked before, so we don't have to
        _check_expression_circular_refs(model, model.metrics[id_], path=path + (id_,))
    for ref in expression.parsed_expr.find_data("dimension"):
        id_ = ref.children[0]
        if id_ in path:
            path = " -> ".join(path + (id_,))
            raise CircularReferenceError(f"Circular reference in calculations: {path}")
        _check_expression_circular_refs(
            model, model.dimensions[id_], path=path + (id_,)
        )


@check_model.depends_on(_check_expr_parse, _check_known_calculations)
def _check_circular_refs(model: Model):
    for expression in model.expressions:
        _check_expression_circular_refs(model, expression)


def _check_measure_uses_aggregate_function(measure: Measure):
    if measure.filter is not None:
        return
    for ref in measure.parsed_expr.find_data("call"):
        fn, *_ = ref.children
        if fn in aggregate_functions:
            return
    raise InvalidMeasureError(
        f"{measure} expression '{measure.str_expr}' only uses other measures "
        "and should be a metric. Remove table from the definition."
    )


@check_model.depends_on(_check_expr_parse, _check_known_functions)
def _check_measures_use_aggregate_functions(model: Model):
    for measure in model.measures.values():
        _check_measure_uses_aggregate_function(measure)


@check_model.depends_on(
    _check_expr_parse,
    _check_known_calculations,
    _check_measure_references_same_table,
    _check_measures_use_aggregate_functions,
)
def _check_measures_dont_reference_filtered_measures(model: Model):
    for measure in model.measures.values():
        for ref in measure.parsed_expr.find_data("measure"):
            id_ = ref.children[0]
            refd = measure.table.measures[id_]
            if refd.filter is not None:
                raise FilteredMeasureReferenceError(
                    f"{measure} references a filtered {refd}, "
                    "filtered measures are not allowed in measure references"
                )


@check_model.depends_on(
    _check_expr_parse, _check_known_calculations, _check_calculation_kinds
)
def _check_aggregate_dimension_table_has_pk(model: Model):
    for dimension in model.dimensions.values():
        if not isinstance(dimension, Expression):
            continue
        for _ in dimension.parsed_expr.find_data("measure"):
            if dimension.table.primary_key is None:
                raise MissingPrimaryKeyError(
                    f"{dimension} is aggregate, so its' parent {dimension.table} "
                    "requires a primary_key"
                )
